# This is a concrete implemention of TaggedValue, an alternative
# representation to the tag_store.py version. This is currently
# only used in tests.

import dataclasses
import typing

from weave.legacy.weave import box
from weave.legacy.weave.language_features.tagging import tag_store


@dataclasses.dataclass
class TaggedValue:
    tag: dict[str, typing.Any]
    value: typing.Any


def concrete_to_tagstore(val: typing.Any) -> typing.Any:
    if isinstance(val, dict):
        return {k: concrete_to_tagstore(v) for k, v in val.items()}
    elif isinstance(val, list):
        return [concrete_to_tagstore(v) for v in val]
    elif isinstance(val, TaggedValue):
        v = box.box(concrete_to_tagstore(val.value))
        tag_store.add_tags(v, concrete_to_tagstore(val.tag))
        return v
    elif dataclasses.is_dataclass(val):
        params = {
            f.name: concrete_to_tagstore(getattr(val, f.name))
            for f in dataclasses.fields(val)
        }
        return val.__class__(**params)
    return val


def concrete_from_tagstore(val: typing.Any) -> typing.Any:
    if tag_store.is_tagged(val):
        return TaggedValue(
            concrete_from_tagstore(tag_store.get_tags(val)),
            _concrete_from_tagstore_inner(val),
        )
    return _concrete_from_tagstore_inner(val)


def _concrete_from_tagstore_inner(val: typing.Any):
    if isinstance(val, dict):
        return {k: concrete_from_tagstore(v) for k, v in val.items()}
    elif isinstance(val, list):
        return [concrete_from_tagstore(v) for v in val]
    elif dataclasses.is_dataclass(val):
        params = {
            f.name: concrete_from_tagstore(getattr(val, f.name))
            for f in dataclasses.fields(val)
        }
        return val.__class__(**params)
    return val



import os
import time

from weave.legacy.weave import cache

# Script to run to delete expired caches
if __name__ == "__main__":
    print("Starting clear cache job", flush=True)
    hour_interval = int(os.getenv("WEAVE_CACHE_CLEAR_INTERVAL", 24))
    print("Clearing expired caches every " + str(hour_interval) + " hours", flush=True)
    while True:
        curTime = time.strftime("%x, %X", time.localtime())
        print("Clearing cache " + curTime, flush=True)
        cache.clear_cache()
        time.sleep(60 * 60 * hour_interval)



import json
import typing

from weave.legacy.weave import environment as weave_env
from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import op
from weave.legacy.weave import file_base, wandb_file_manager
from weave.legacy.weave.artifact_fs import FilesystemArtifactDir, FilesystemArtifactFile
from weave.legacy.weave.artifact_wandb import WandbArtifact, WandbArtifactManifest


@op(name="dir-pathReturnType", hidden=True)
def path_return_type(dir: file_base.Dir, path: str) -> types.Type:
    return types.TypeRegistry.type_of(dir.path_info(path))


@op(
    name="dir-path",
    refine_output_type=path_return_type,
)
def open(
    dir: file_base.Dir, path: str
) -> typing.Union[file_base.File, file_base.Dir, None]:
    return dir.path_info(path)


def _file_dict_from_manifest(
    file: FilesystemArtifactFile, manifest: WandbArtifactManifest
) -> dict:
    art = file.artifact
    if not isinstance(art, WandbArtifact):
        return _file_dict(file)

    entry = manifest.get_entry_by_path(file.path)
    if entry is None:
        return _file_dict(file)

    digest = entry.get("digest")
    size = entry.get("size") or 0

    base_dict = {
        "birthArtifactID": "TODO",
        "digest": digest,
        "fullPath": file.path,
        "size": size,
        "type": "file",
    }

    ref = entry.get("ref")
    if ref is not None:
        base_dict["ref"] = ref

    # if the ref exists then use that for the URL
    if "ref" in base_dict:
        base_dict["url"] = base_dict["ref"]
    else:
        res = wandb_file_manager._local_path_and_download_url(
            art.uri_obj.with_path(file.path),
            manifest,
            base_url=weave_env.wandb_frontend_base_url(),
        )
        if res:
            base_dict["url"] = res[1]
        else:
            base_dict["url"] = None

    return base_dict


def _file_dict(file: FilesystemArtifactFile) -> dict:
    if isinstance(file.artifact, WandbArtifact):
        raise ValueError(
            "WandbArtifact passed to _file_dict. This is not supported. Use _file_dict_from_manifest instead."
        )

    return {
        "birthArtifactID": "TODO",
        "digest": file.digest(),
        "fullPath": file.path,
        "size": file.artifact.size(file.path),
        "type": "file",
        "url": file.artifact.direct_url(file.path),
    }


@op(
    name="dir-_as_w0_dict_",
)
def _as_w0_dict_(dir: file_base.Dir) -> dict:
    if isinstance(dir, FilesystemArtifactDir) and isinstance(
        dir.artifact, WandbArtifact
    ):
        manifest = dir.artifact._manifest()
    else:
        manifest = None
    if not isinstance(dir, FilesystemArtifactDir):
        raise errors.WeaveInternalError("Dir must be FilesystemArtifactDir")
    return {
        "fullPath": dir.fullPath,
        "size": dir.size,
        "dirs": {
            k: {
                "fullPath": v.fullPath,
                "size": v.size,
                "dirs": v.dirs,
                "files": v.files,
            }
            for k, v in dir.dirs.items()
        },
        "files": {
            k: _file_dict_from_manifest(v, manifest)
            if manifest is not None
            else _file_dict(v)
            for k, v in dir.files.items()
        },
    }


@op(name="file-directUrlAsOf")
def direct_url_as_of(file: file_base.File, asOf: int) -> str:
    # This should only be used in the first dispatch phase
    raise NotImplementedError


@op(name="file-contents")
def file_contents(file: file_base.File) -> str:
    with file.open("r") as f:
        return f.read()


@op(name="file-dir")
def artifactfile_dir(
    item: typing.Union[file_base.File, file_base.Dir],
) -> typing.Optional[file_base.Dir]:
    if not isinstance(item, file_base.Dir):
        return None
    return item


@op(name="file-type")
def file_type(file: typing.Union[file_base.File, file_base.Dir]) -> types.Type:
    return types.TypeRegistry.type_of(file)


@op(name="file-path")
def file_path(file: file_base.File) -> str:
    return file.path


@op(name="file-size")
def file_size(file: file_base.File) -> int:
    return file.size()


@op(name="file-digest")
def file_digest(file: file_base.File) -> typing.Optional[str]:
    return file.digest()



import pytest

import weave
from weave.legacy.weave import (
    artifact_fs,
    artifact_local,
    environment,
    errors,
    storage,
    wandb_api,
)


@pytest.fixture()
def public_env():
    orig_is_public = environment.is_public
    environment.is_public = lambda: True
    token = wandb_api.set_wandb_api_context("test-user-id", None, None, None)
    try:
        yield
    finally:
        environment.is_public = orig_is_public
        wandb_api.reset_wandb_api_context(token)


def test_access_file(public_env):
    with pytest.raises(errors.WeaveAccessDeniedError):
        weave.use(weave.legacy.weave.ops.local_path("/tmp/bad.json"))


@pytest.mark.parametrize("path", ["..", "/tmp", "//tmp", "//tmp/bad.json", "/tmp/.../"])
def test_access_artifact(public_env, path):
    ref = storage.save(5)
    with pytest.raises(errors.WeaveAccessDeniedError):
        info = ref.artifact.path_info(path)

    info = artifact_fs.FilesystemArtifactFile(ref.artifact, path)
    with pytest.raises((errors.WeaveAccessDeniedError, IsADirectoryError)):
        with info.open() as f:
            pass

    art = artifact_local.LocalArtifact("test-artifact", None)
    with pytest.raises(errors.WeaveAccessDeniedError):
        with art.new_dir(path) as d:
            pass

    with pytest.raises((errors.WeaveAccessDeniedError, IsADirectoryError)):
        with art.new_file(path) as d:
            pass



import itertools

import numpy as np
import pytest

from weave.legacy.weave import api as weave
from weave.legacy.weave import box, ops, weave_internal
from weave.legacy.weave import ops_arrow as arrow
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.language_features.tagging import (
    make_tag_getter_op,
    tag_store,
    tagged_value_type,
)
from weave.legacy.weave.ops_primitives import dict_, list_

from weave.legacy.tests.util import tag_test_util as ttu
from weave.legacy.tests.util import list_arrow_test_helpers as lath


def filter_fn(row) -> bool:
    return row < 3


def inv_filter_fn(row) -> bool:
    return row < -3


# TODO: Test Tag managmenet
@pytest.mark.parametrize(
    "data, fn_name, fn_def, res, extra_args",
    [
        ([1, 2, 3, 4], "map", lambda row: row + 1, [2, 3, 4, 5], []),
        (
            [1, 2, 3, 4],
            "map",
            lambda row: list_.make_list(a=row + 1, b=row * -1),
            [[2, -1], [3, -2], [4, -3], [5, -4]],
            [],
        ),
        (
            [1, 2, 3, 4],
            "map",
            lambda row: dict_(a=row + 1, b=row * -1),
            [
                {"a": 2, "b": -1},
                {"a": 3, "b": -2},
                {"a": 4, "b": -3},
                {"a": 5, "b": -4},
            ],
            [],
        ),
        ([1, 2, 3, 4], "filter", filter_fn, [1, 2], []),
        ([1, None, 2, None, 3, None, 4], "filter", filter_fn, [1, 2], []),
        ([1, 2, 3, 4], "filter", inv_filter_fn, [], []),
        (
            [1, 2, 3, 4],
            "sort",
            lambda row: list_.make_list(a=row * -1),
            [4, 3, 2, 1],
            [["asc"]],
        ),
        (
            [1, 2, 3, 4],
            "sort",
            lambda row: list_.make_list(a=row * -1),
            [1, 2, 3, 4],
            [["desc"]],
        ),
        (
            [1, 2, 3, 4],
            "sort",
            lambda row: list_.make_list(a=row * 0, b=row * -2),
            [4, 3, 2, 1],
            [["asc", "asc"]],
        ),
        (
            [1, 2, 3, 4],
            "sort",
            lambda row: list_.make_list(a=row * 0, b=row * -2),
            [1, 2, 3, 4],
            [["asc", "desc"]],
        ),
        # ([1,2,3,4], "map", lambda row: row + 1, [2,3,4,5]),
        (
            [1, 2, 3, 4, None],
            "sort",
            lambda row: list_.make_list(
                a=row * 0 if row is not None else None,
                b=row * -2 if row is not None else None,
            ),
            [None, 1, 2, 3, 4],
            [["asc", "desc"]],
        ),
        (
            [1, 2, 3, 4, None],
            "sort",
            lambda row: list_.make_list(
                a=row * 0 if row is not None else None,
                b=row * -2 if row is not None else None,
            ),
            [None, 4, 3, 2, 1],
            [["asc", "asc"]],
        ),
        (
            [1, 2, 3, 4, None],
            "sort",
            lambda row: list_.make_list(
                a=row * 0 if row is not None else None,
                b=row * -2 if row is not None else None,
            ),
            [4, 3, 2, 1, None],
            [["desc", "asc"]],
        ),
        (
            [1, 2, 3, 4, None],
            "sort",
            lambda row: list_.make_list(a="a", b=row),
            [None, 1, 2, 3, 4],
            [["asc", "asc"]],
        ),
        (
            [[None, 1], [None, 2], [None, 3]],
            "sort",
            lambda row: list_.make_list(a=row[0], b=row[1]),
            [[None, 1], [None, 2], [None, 3]],
            [["asc", "asc"]],
        ),
        (
            [1, "a", None, 2.0, {"a": 1}],
            "sort",
            lambda row: list_.make_list(a=row),
            [None, 1, 2.0, "a", {"a": 1}],
            [["asc"]],
        ),
        (
            [{"a": 1}, {"a": 2}, {"a": 3}, {"a": None}],
            "sort",
            lambda row: list_.make_list(a=row["a"]),
            list(
                reversed(
                    [
                        {"a": None},
                        {"a": 1},
                        {"a": 2},
                        {"a": 3},
                    ]
                )
            ),
            [["desc"]],
        ),
    ],
)
def test_list_arrow_compatibility(data, fn_name, fn_def, res, extra_args):
    test_list_arrow_fn_results_maybe_incompatible(
        data, fn_name, fn_def, res, res, extra_args
    )


@pytest.mark.parametrize(
    "data, fn_name, fn_def, list_res, arrow_res, extra_args",
    [
        (
            [{"a": 1}, {"a": 2}, {"a": 3}, {"a": None}],
            "sort",
            lambda row: list_.make_list(a=row),
            [{"a": 1}, {"a": 2}, {"a": 3}, {"a": None}],
            [
                {"a": None},
                {"a": 1},
                {"a": 2},
                {"a": 3},
            ],
            [["asc"]],
        )
    ],
)
def test_list_arrow_fn_results_maybe_incompatible(
    data, fn_name, fn_def, list_res, arrow_res, extra_args
):
    list_node = list_.make_list(**{f"{i}": v for i, v in enumerate(data)})
    arrow_node = weave.save(arrow.to_arrow(data))

    list_node_mapped = list_node.__getattr__(fn_name)(fn_def, *extra_args)
    arrow_node_mapped = arrow_node.__getattr__(fn_name)(fn_def, *extra_args)

    assert weave.use(list_node_mapped) == list_res
    assert weave.use(arrow_node_mapped).to_pylist_raw() == arrow_res


# TODO: In Weave0, we also have a `join2` op. This will be done in a followup
# PR.
def compare_join_results(results, exp_results):
    # Currently list join and arrow join return slightly different result
    # orderings. PyArrow's join pushes all outer join results without a match to
    # the end of the list, while list join returns the order found. For purposes
    # of Weave1 development, we will ignore this difference. It would be
    # unnecessarily costly to resort the results of either join to match the
    # other.
    error_msg = f"Expected {exp_results}, got {results}"
    assert len(results) == len(exp_results), f"Result length is wrong; {error_msg}"
    for exp_result in exp_results:
        assert (
            exp_result in results
        ), f"Expected result {exp_result} not found; {error_msg}"
    for result in results:
        assert (
            result in exp_results
        ), f"unexpected result {exp_result} found; {error_msg}"


@pytest.mark.parametrize("li", lath.ListInterfaces)
def test_join_all(li):
    # This test tests joining a nullable list of lists, with duplicate keys that are partially overlapping and not all entries have the key! (our join logic is quite nuanced.)
    list_node_1 = li.make_node(
        [
            {
                "id": "1.0",
            },
            {
                "id": "1.a",
                "val": 1,
            },
            {
                "id": "1.b",
                "val": 1,
            },
            {
                "id": "1.c",
                "val": 2,
            },
            {
                "id": "1.d",
                "val": 2,
            },
            {
                "id": "1.e",
                "val": 3,
            },
        ]
    )
    list_node_2 = li.make_node(
        [
            {
                "id": "2.0",
            },
            {
                "id": "2.f",
                "val": 1,
            },
            {
                "id": "2.g",
                "val": 1,
            },
            {
                "id": "2.h",
                "val": 3,
            },
            {
                "id": "2.i",
                "val": 3,
            },
            {
                "id": "2.j",
                "val": 5,
            },
        ]
    )
    join_list = list_.make_list(a=list_node_1, c=None, b=list_node_2)
    join_fn = weave_internal.define_fn(
        {"row": list_node_1.type.object_type},
        lambda row: row["val"],
    )

    joined_inner_node = join_list.joinAll(join_fn, False)
    joined_outer_node = join_list.joinAll(join_fn, True)

    # TODO: Arrow and List have different permutation ordering here - probably fix list implementation to match arrow
    exp_results = [
        {
            "val": [1, 1],
            "id": ["1.a", "2.f"],
        },
        {
            "val": [1, 1],
            "id": ["1.a", "2.g"],
        },
        {
            "val": [1, 1],
            "id": ["1.b", "2.f"],
        },
        {
            "val": [1, 1],
            "id": ["1.b", "2.g"],
        },
        {
            "val": [3, 3],
            "id": ["1.e", "2.i"],
        },
        {
            "val": [3, 3],
            "id": ["1.e", "2.h"],
        },
    ]
    compare_join_results(li.use_node(joined_inner_node), exp_results)

    exp_results = [
        {
            "val": [1, 1],
            "id": ["1.a", "2.f"],
        },
        {
            "val": [1, 1],
            "id": ["1.b", "2.f"],
        },
        {
            "val": [1, 1],
            "id": ["1.a", "2.g"],
        },
        {
            "val": [1, 1],
            "id": ["1.b", "2.g"],
        },
        {
            "val": [2, None],
            "id": ["1.c", None],
        },
        {
            "val": [2, None],
            "id": ["1.d", None],
        },
        {
            "val": [3, 3],
            "id": ["1.e", "2.h"],
        },
        {
            "val": [3, 3],
            "id": ["1.e", "2.i"],
        },
        {
            "val": [None, 5],
            "id": [None, "2.j"],
        },
    ]

    compare_join_results(li.use_node(joined_outer_node), exp_results)
    # Currently list join and arrow join return slightly different result
    # orderings. DuckDB's join order is not obvious, while
    # list join returns the order found. For purposes
    # of Weave1 development, we will ignore this difference. It would be
    # unnecessarily costly to resort the results of either join to match the
    # other.
    if li == lath.ListNode:
        tag_order = [1, 1, 1, 1, 2, 2, 3, 3, 5]
    elif li == lath.ArrowNode:
        tag_order = [1, 1, 1, 1, 3, 3, 2, 2, 5]
    assert li.use_node(joined_outer_node.joinObj()) == tag_order


@pytest.mark.parametrize("li", lath.ListInterfaces)
def test_join_2(li):
    # This test tests joining a nullable list of lists, with duplicate keys that are partially overlapping and not all entries have the key! (our join logic is quite nuanced.)
    list_node_1 = li.make_node(
        [
            {
                "id": "1.0",
            },
            {
                "id": "1.a",
                "val": 1,
            },
            {
                "id": "1.b",
                "val": 1,
            },
            {
                "id": "1.c",
                "val": 2,
            },
            {
                "id": "1.d",
                "val": 2,
            },
            {
                "id": "1.e",
                "val": 3,
            },
        ]
    )
    list_node_2 = li.make_node(
        [
            {
                "id": "2.0",
            },
            {
                "id": "2.f",
                "val": 1,
            },
            {
                "id": "2.g",
                "val": 1,
            },
            {
                "id": "2.h",
                "val": 3,
            },
            {
                "id": "2.i",
                "val": 3,
            },
            {
                "id": "2.j",
                "val": 5,
            },
        ]
    )
    # join_list = list_.make_list(a=list_node_1, c=None, b=list_node_2)
    join_fn = weave_internal.define_fn(
        {"row": list_node_1.type.object_type},
        lambda row: row["val"],
    )

    joined_inner_node = list_node_1.join(
        list_node_2, join_fn, join_fn, "a0", "a1", False, False
    )
    joined_left_outer_node = list_node_1.join(
        list_node_2, join_fn, join_fn, "a0", "a1", True, False
    )
    joined_right_outer_node = list_node_1.join(
        list_node_2, join_fn, join_fn, "a0", "a1", False, True
    )
    joined_full_outer_node = list_node_1.join(
        list_node_2, join_fn, join_fn, "a0", "a1", True, True
    )

    # TODO: Arrow and List have different permutation ordering here - probably fix list implementation to match arrow
    exp_results = [
        {"a0": {"id": "1.a", "val": 1}, "a1": {"id": "2.g", "val": 1}},
        {"a0": {"id": "1.b", "val": 1}, "a1": {"id": "2.g", "val": 1}},
        {"a0": {"id": "1.e", "val": 3}, "a1": {"id": "2.i", "val": 3}},
        {"a0": {"id": "1.a", "val": 1}, "a1": {"id": "2.f", "val": 1}},
        {"a0": {"id": "1.b", "val": 1}, "a1": {"id": "2.f", "val": 1}},
        {"a0": {"id": "1.e", "val": 3}, "a1": {"id": "2.h", "val": 3}},
    ]
    compare_join_results(li.use_node(joined_inner_node), exp_results)

    exp_results = [
        {"a0": {"id": "1.a", "val": 1}, "a1": {"id": "2.g", "val": 1}},
        {"a0": {"id": "1.b", "val": 1}, "a1": {"id": "2.g", "val": 1}},
        {"a0": {"id": "1.e", "val": 3}, "a1": {"id": "2.i", "val": 3}},
        {"a0": {"id": "1.a", "val": 1}, "a1": {"id": "2.f", "val": 1}},
        {"a0": {"id": "1.b", "val": 1}, "a1": {"id": "2.f", "val": 1}},
        {"a0": {"id": "1.e", "val": 3}, "a1": {"id": "2.h", "val": 3}},
        {"a0": {"id": "1.c", "val": 2}, "a1": None},
        {"a0": {"id": "1.d", "val": 2}, "a1": None},
    ]
    compare_join_results(li.use_node(joined_left_outer_node), exp_results)

    exp_results = [
        {"a0": {"id": "1.b", "val": 1}, "a1": {"id": "2.f", "val": 1}},
        {"a0": {"id": "1.b", "val": 1}, "a1": {"id": "2.g", "val": 1}},
        {"a0": {"id": "1.e", "val": 3}, "a1": {"id": "2.h", "val": 3}},
        {"a0": {"id": "1.e", "val": 3}, "a1": {"id": "2.i", "val": 3}},
        {"a0": {"id": "1.a", "val": 1}, "a1": {"id": "2.f", "val": 1}},
        {"a0": {"id": "1.a", "val": 1}, "a1": {"id": "2.g", "val": 1}},
        {"a0": None, "a1": {"id": "2.j", "val": 5}},
    ]
    compare_join_results(li.use_node(joined_right_outer_node), exp_results)

    exp_results = [
        {"a0": {"id": "1.a", "val": 1}, "a1": {"id": "2.g", "val": 1}},
        {"a0": {"id": "1.b", "val": 1}, "a1": {"id": "2.g", "val": 1}},
        {"a0": {"id": "1.e", "val": 3}, "a1": {"id": "2.i", "val": 3}},
        {"a0": {"id": "1.a", "val": 1}, "a1": {"id": "2.f", "val": 1}},
        {"a0": {"id": "1.b", "val": 1}, "a1": {"id": "2.f", "val": 1}},
        {"a0": {"id": "1.e", "val": 3}, "a1": {"id": "2.h", "val": 3}},
        {"a0": {"id": "1.c", "val": 2}, "a1": None},
        {"a0": {"id": "1.d", "val": 2}, "a1": None},
        {"a0": None, "a1": {"id": "2.j", "val": 5}},
    ]
    compare_join_results(li.use_node(joined_full_outer_node), exp_results)

    # Currently list join and arrow join return slightly different result
    # orderings. DuckDB's join order is not obvious, while
    # list join returns the order found. For purposes
    # of Weave1 development, we will ignore this difference. It would be
    # unnecessarily costly to resort the results of either join to match the
    # other.
    if li == lath.ListNode:
        tag_order = [1, 1, 1, 1, 2, 2, 3, 3, 5]
    elif li == lath.ArrowNode:
        tag_order = [1, 1, 1, 1, 3, 3, 2, 2, 5]
    assert li.use_node(joined_full_outer_node.joinObj()) == tag_order


algos = [
    ("pca", {}),
    ("tsne", {"perplexity": 2, "learning_rate": 10, "iterations": 3}),
    ("umap", {"neighbors": 2, "min_dist": 0.1, "spread": 1.0}),
]


@pytest.mark.timeout(60)
@pytest.mark.parametrize(
    "li, algo, options",
    [
        (li, algo_options[0], algo_options[1])
        for li in lath.ListInterfaces
        for algo_options in algos
    ],
)
def test_2d_projection(li, algo, options):
    n_rows = 15
    n_cols = 6
    data = np.random.rand(n_rows, n_cols)
    data_as_dicts = [
        {f"col_{item_ndx}": item for item_ndx, item in enumerate(row)} for row in data
    ]
    col_names = [f"col_{item_ndx}" for item_ndx in range(n_cols)]
    node = li.make_node(data_as_dicts)
    projection = node._get_op("2DProjection")(algo, "many", col_names, options)
    res = li.use_node(projection)
    assert len(res) == n_rows
    assert res[0].get("projection").get("x") is not None
    assert res[0].get("projection").get("y") is not None
    assert res[0].get("source") == data_as_dicts[0]


def test_join_to_str():
    data = ["1", None, "2", None, "3"]
    res = "1,,2,,3"  # note, Weave0 uses "" for Nones - we may consider changing this in the future
    list_node = lath.ListNode.make_node(data)
    arrow_node = lath.ArrowNode.make_node([data, data])

    list_joined = list_node.joinToStr(",")
    arrow_joined = arrow_node.joinToStr(",")

    assert lath.ListNode.use_node(list_joined) == res
    assert lath.ArrowNode.use_node(arrow_joined) == [res, res]


@pytest.mark.parametrize(
    "use_arrow",
    [True, False],
)
def test_flatten_and_tags(use_arrow):
    # tagged<list<tagged<awl<tagged<dict<id: tagged<number>, col: tagged<string>>>>>>>
    cst = weave_internal.const
    tag = ttu.op_add_tag
    item_tag_getter = ttu.make_get_tag("item_tag")
    list_tag_getter = ttu.make_get_tag("list_tag")

    def cst_list(data):
        return ops.make_list(**{str(i): l for i, l in enumerate(data)})

    def cst_arrow(data):
        return arrow.ops.list_to_arrow(cst_list(data))

    if use_arrow:
        inner_list_fn = cst_arrow
    else:
        inner_list_fn = cst_list

    list_of_lists = cst_list(
        [
            tag(
                inner_list_fn(
                    [
                        tag(cst("item_1_1"), {"item_tag": "1_1"}),
                        tag(cst("item_1_2"), {"item_tag": "1_2"}),
                        tag(cst("item_1_3"), {"item_tag": "1_3"}),
                    ]
                ),
                {"list_tag": "1"},
            ),
            tag(
                inner_list_fn(
                    [
                        tag(cst("item_2_1"), {"item_tag": "2_1"}),
                        tag(cst("item_2_2"), {"item_tag": "2_2"}),
                        tag(cst("item_2_3"), {"item_tag": "2_3"}),
                    ]
                ),
                {"list_tag": "2"},
            ),
            tag(
                inner_list_fn(
                    [
                        tag(cst("item_3_1"), {"item_tag": "3_1"}),
                        tag(cst("item_3_2"), {"item_tag": "3_2"}),
                        tag(cst("item_3_3"), {"item_tag": "3_3"}),
                    ]
                ),
                {"list_tag": "3"},
            ),
        ]
    )

    flattened = list_of_lists.flatten()

    data_res = weave.use(flattened)
    item_tag_res = weave.use(item_tag_getter(flattened[0]))
    list_tag_res = weave.use(list_tag_getter(flattened[0]))

    if use_arrow:
        data_res = data_res.to_pylist_notags()

    assert data_res == [
        "item_1_1",
        "item_1_2",
        "item_1_3",
        "item_2_1",
        "item_2_2",
        "item_2_3",
        "item_3_1",
        "item_3_2",
        "item_3_3",
    ]
    assert item_tag_res == "1_1"
    assert list_tag_res == "1"


@pytest.mark.parametrize(
    "use_arrow",
    [True, False],
)
def test_tag_pushdown_on_list_of_lists(use_arrow):
    data = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    for i, row in enumerate(data):
        for j, item in enumerate(row):
            data[i][j] = tag_store.add_tags(box.box(item), {"row": i, "col": j})
        data[i] = tag_store.add_tags(box.box(data[i]), {"list_tag": i})
    data = tag_store.add_tags(box.box(data), {"top_tag": 2})
    list_node = weave.save(data)

    if use_arrow:
        list_node = arrow.ops.list_to_arrow(list_node)

    from weave.legacy.weave import context_state

    _loading_builtins_token = context_state.set_loading_built_ins()

    row_tag_getter = make_tag_getter_op.make_tag_getter_op("row", types.Int())
    col_tag_getter = make_tag_getter_op.make_tag_getter_op("col", types.Int())
    list_tag_getter = make_tag_getter_op.make_tag_getter_op("list_tag", types.Int())
    top_tag_getter = make_tag_getter_op.make_tag_getter_op("top_tag", types.Int())

    context_state.clear_loading_built_ins(_loading_builtins_token)

    inner_func = (
        lambda n: n
        + row_tag_getter(n)
        + col_tag_getter(n)
        + list_tag_getter(n)
        - top_tag_getter(n)
        + 3
    )

    outer_func = lambda row: row.map(inner_func)
    mapped = list_node.map(outer_func)

    assert mapped.type == tagged_value_type.TaggedValueType(
        types.TypedDict({"top_tag": types.Int()}),
        (types.List if not use_arrow else arrow.ArrowWeaveListType)(
            tagged_value_type.TaggedValueType(
                types.TypedDict({"list_tag": types.Int(), "top_tag": types.Int()}),
                types.List(
                    tagged_value_type.TaggedValueType(
                        types.TypedDict(
                            {
                                "row": types.Int(),
                                "col": types.Int(),
                                "top_tag": types.Int(),
                                "list_tag": types.Int(),
                            }
                        ),
                        types.Number(),
                    )
                ),
            )
        ),
    )

    output = weave.use(mapped)

    if use_arrow:
        output = output.to_pylist_tagged()

    assert output == [
        [data[i][j] + 3 + i + j - 2 + i for j in range(3)] for i in range(3)
    ]


@pytest.mark.parametrize(
    "use_arrow",
    [True, False],
)
def test_sample(use_arrow):
    input_array = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    if use_arrow:
        input_array_node = weave.save(arrow.to_arrow(input_array))
        sample_node = input_array_node.randomlyDownsample(5)
    else:
        sample_node = list_.sample(input_array, 5)
    result = weave.use(sample_node)

    if use_arrow:
        result = result.to_pylist_tagged()

    assert len(result) == 5
    assert len(set(result)) == 5
    assert set(result).issubset(set(input_array))
    assert sorted(result) == result


@pytest.mark.parametrize(
    "use_arrow",
    [True, False],
)
def test_sample_10(use_arrow):
    input_array = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    if use_arrow:
        input_array_node = weave.save(arrow.to_arrow(input_array))
        sample_node_10 = input_array_node.randomlyDownsample(10)
    else:
        sample_node_10 = list_.sample(input_array, 10)
    result_10 = weave.use(sample_node_10)

    if use_arrow:
        result_10 = result_10.to_pylist_tagged()

    assert result_10 == input_array


@pytest.mark.parametrize(
    "use_arrow",
    [True, False],
)
def test_sample_bad(use_arrow):
    input_array = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    if use_arrow:
        input_array_node = weave.save(arrow.to_arrow(input_array))
        bad_node = input_array_node.randomlyDownsample(-1)
    else:
        bad_node = list_.sample(input_array, -1)

    with pytest.raises(ValueError):
        weave.use(bad_node)


@pytest.mark.parametrize(
    "use_arrow, input_array",
    list(itertools.product([True, False], [[], list(range(10))])),
)
def test_sample_length_zero_list(use_arrow, input_array):
    input_array = []
    if use_arrow:
        input_array_node = weave.save(arrow.to_arrow(input_array))
        bad_node = input_array_node.randomlyDownsample(0)
    else:
        bad_node = list_.sample(input_array, 0)

    result = weave.use(bad_node)

    if use_arrow:
        result = result.to_pylist_tagged()

    assert result == []



from weave.legacy.weave import weave_types as types
from weave.legacy.weave import decorator_op
from weave.legacy.weave.language_features.tagging import tagging_op_logic


@decorator_op.op(
    name="op_get_tag_type",
    input_type={
        "obj_type": types.TypeType(),
    },
    output_type=types.TypeType(),
)
def op_get_tag_type(obj_type):  # type: ignore
    return tagging_op_logic.op_get_tag_type_resolver(obj_type)


@decorator_op.op(
    name="op_make_type_tagged",
    input_type={
        "obj_type": types.TypeType(),
        "tag_type": types.TypeType(),
    },
    output_type=types.TypeType(),
)
def op_make_type_tagged(obj_type, tag_type):  # type: ignore
    return tagging_op_logic.op_make_type_tagged_resolver(obj_type)


@decorator_op.op(
    name="op_make_type_key_tag",
    input_type={
        "obj_type": types.TypeType(),
        "key": types.String(),
        "tag_type": types.TypeType(),
    },
    output_type=types.TypeType(),
)
def op_make_type_key_tag(obj_type, key, tag_type):  # type: ignore
    return tagging_op_logic.op_make_type_key_tag_resolver(obj_type, key, tag_type)



from typing import Iterable, Optional


class WeaveUnmergableArtifactsError(Exception):
    pass


class WeaveFingerprintErrorMixin:
    fingerprint: Optional[Iterable] = None


class WeaveBaseError(Exception, WeaveFingerprintErrorMixin):
    pass


class WeaveBaseWarning(Warning):
    pass


# Only use this if you actually want to return an Http 400
# to the client. This should only happen in cases where the
# client is wrong.
class WeaveBadRequest(WeaveBaseError):
    pass


class WeaveInternalError(WeaveBaseError):
    """Internal Weave Error (a programming error)"""

    pass


class WeaveSerializeError(WeaveBaseError):
    pass


class WeaveConfigurationError(WeaveBaseError):
    pass


class WeaveApiError(WeaveBaseError):
    pass


class WeaveTypeError(WeaveBaseError):
    pass


class WeaveTypeWarning(WeaveBaseWarning):
    pass


class WeaveDefinitionError(WeaveBaseError):
    pass


class WeaveMakeFunctionError(Exception):
    pass


class WeaveExpectedConstError(Exception):
    pass


class WeaveInvalidURIError(Exception):
    pass


class WeaveStorageError(Exception):
    pass


class WeaveExecutionError(Exception):
    pass


class WeavifyError(Exception):
    pass


class WeaveDispatchError(WeaveBaseError):
    pass


class WeaveArtifactFileNotFound(WeaveBaseError):
    pass


class WeaveArtifactCollectionNotFound(WeaveBaseError):
    pass


class WeaveArtifactVersionNotFound(WeaveBaseError):
    pass


class WeaveVectorizationError(WeaveBaseError):
    pass


class WeaveValueError(WeaveBaseError):
    pass


class WeaveClientArtifactResolutionFailure(WeaveBaseError):
    pass


class WeaveTableDeserializationError(WeaveBaseError):
    pass


class WeaveStitchGraphMergeError(WeaveBaseError):
    pass


class WeaveHashConstTypeError(WeaveBaseError):
    """Raised if __hash__ is called on a Const Type.

    To hash a Const Type, we'd need to hash the value, which is unbounded.
    """

    pass


class WeaveGQLCompileError(WeaveBaseError):
    pass


class WeaveGQLExecuteMissingAliasError(WeaveBaseError):
    pass


class WeaveAccessDeniedError(WeaveBaseError):
    pass


class WeaveWandbArtifactManagerError(WeaveBaseError):
    pass


class WeaveArtifactMediaFileLookupError(WeaveBaseError):
    pass


class WeaveClientRequestError(WeaveBaseError):
    pass


class WeaveMissingVariableError(WeaveBaseError):
    pass


class WeaveWBHistoryTranslationError(WeaveBaseError):
    pass


class WeaveWandbAuthenticationException(Exception):
    pass


class WeaveMissingOpDefError(WeaveBaseError):
    pass


class WeaveMergeArtifactSpecError(WeaveBaseError):
    pass


class WeaveOpSerializeError(WeaveBaseError):
    pass


class WeaveInitError(WeaveBaseError):
    pass



import dataclasses
import typing

import weave
from weave.legacy.weave import graph, panel

ExpressionType = typing.TypeVar("ExpressionType")


@weave.type()
class GroupingEditorConfig(typing.Generic[ExpressionType]):
    node: graph.Node[ExpressionType] = dataclasses.field(default_factory=graph.VoidNode)


@weave.type()
class GroupingEditor(panel.Panel):
    id = "GroupingEditor"
    config: GroupingEditorConfig = dataclasses.field(
        default_factory=GroupingEditorConfig
    )

    def __init__(
        self, input_node=graph.VoidNode(), vars=None, config=None, **options
    ) -> None:
        super().__init__(input_node=input_node, vars=vars)
        self.config = config
        if self.config is None:
            self.config = GroupingEditorConfig()
        if "node" in options:
            self.config.node = options["node"]



import time

import wandb

import weave
from weave.legacy.weave.ecosystem.wandb.panel_time_series import TimeSeries


def test_panel_timeseries(user_by_api_key_in_env):
    run = wandb.init(project="project_exists")
    for i in range(10):
        time.sleep(0.2)
        run.log({"val": i, "cat": str(i % 2)})
    run.finish()

    history_node = (
        weave.legacy.weave.ops.project(run.entity, run.project).run(run.id).history2()
    )
    panel = TimeSeries(history_node)
    init_config_node = panel.initialize()
    init_config = weave.use(init_config_node)
    panel.config = init_config
    render_node = panel.render()
    res = weave.use(render_node)
    # What to assert here? Should we be getting .contents?
    assert res != None



import weave
from weave.legacy.weave import graph, ops, panel, panel_util


@weave.type()
class PanelMarkdown(panel.Panel):
    id = "markdown"

    def __init__(self, input_node=graph.VoidNode(), vars=None, config=None, **options):
        # Convert input_node here
        # TODO: Type adaptation should be done by render, not
        #     construction (ie the Panel should handle Markdown and File<Markdown>
        #     types)
        input_node = panel_util.make_node(input_node)
        if ops.MarkdownType().assign_type(input_node.type):
            input_node = ops.markdown_file(input_node)
        super().__init__(input_node=input_node, vars=vars)



import wandb

import weave
from weave.legacy.weave import artifact_local, artifact_wandb
from weave.legacy.weave.wandb_interface.wandb_artifact_pusher import (
    write_artifact_to_wandb,
)


def without_keys(d, keys):
    return {k: v for k, v in d.items() if k not in keys}


def test_artifact_metadata(user_by_api_key_in_env):
    # Write some metadata to a local artifact
    local_art = artifact_local.LocalArtifact("test_artifact")
    local_art.set("obj", weave.types.Number(), 1)
    assert local_art.metadata.as_dict() == {}
    local_art.metadata["k_1"] = "v_1"
    local_art.metadata["k_2"] = "v_2"
    assert local_art.metadata.as_dict() == {"k_1": "v_1", "k_2": "v_2"}
    local_art.save()
    assert local_art.metadata.as_dict() == {
        "k_1": "v_1",
        "k_2": "v_2",
    }

    # Load the artifact and check the metadata
    local_art = artifact_local.LocalArtifact("test_artifact", "latest")
    assert local_art.metadata.as_dict() == {
        "k_1": "v_1",
        "k_2": "v_2",
    }

    # Overwrite one of the keys and add a new one
    local_art.metadata["k_1"] = "v_3"
    local_art.metadata["k_3"] = "v_4"
    assert local_art.metadata.as_dict() == {
        "k_1": "v_3",
        "k_2": "v_2",
        "k_3": "v_4",
    }
    local_art.save()
    assert local_art.metadata.as_dict() == {
        "k_1": "v_3",
        "k_2": "v_2",
        "k_3": "v_4",
    }

    # Verify that the metadata is correct
    local_art = artifact_local.LocalArtifact("test_artifact", "latest")
    assert local_art.metadata.as_dict() == {
        "k_1": "v_3",
        "k_2": "v_2",
        "k_3": "v_4",
    }

    # Push an artifact to wandb and verify that the metadata is correct
    remote_uri = weave.legacy.weave.ops.publish_artifact(
        weave.legacy.weave.ops.get(local_art.uri + "/obj"),
        "test_artifact",
        "test_project",
        None,
    )
    remote_ref = artifact_wandb.WandbArtifactRef.from_uri(
        artifact_wandb.WeaveWBArtifactURI.parse(remote_uri)
    )
    remote_art = remote_ref.artifact
    assert without_keys(remote_art.metadata.as_dict(), ["_weave_meta"]) == {
        "k_1": "v_3",
        "k_2": "v_2",
        "k_3": "v_4",
    }

    # Fork the artifact and sure the metadata persists
    forked_art = artifact_local.LocalArtifact.fork_from_uri(
        artifact_wandb.WeaveWBArtifactURI.parse(remote_uri)
    )
    assert without_keys(forked_art.metadata.as_dict(), ["_weave_meta"]) == {
        "k_1": "v_3",
        "k_2": "v_2",
        "k_3": "v_4",
    }
    forked_art.metadata["k_1"] = "v_5"
    forked_art.metadata["k_4"] = "v_6"
    assert without_keys(forked_art.metadata.as_dict(), ["_weave_meta"]) == {
        "k_1": "v_5",
        "k_2": "v_2",
        "k_3": "v_4",
        "k_4": "v_6",
    }
    forked_art.save()
    assert without_keys(forked_art.metadata.as_dict(), ["_weave_meta"]) == {
        "k_1": "v_5",
        "k_2": "v_2",
        "k_3": "v_4",
        "k_4": "v_6",
    }


def test_artifact_files_count(user_by_api_key_in_env):
    run = wandb.init(project="project_exists")
    artifact = wandb.Artifact("test", "datatest")
    table0 = wandb.Table(data=[[1, 2, 3]], columns=["a", "b", "c"])
    table1 = wandb.Table(data=[[1, 2, 3]], columns=["a", "b", "c"])
    table2 = wandb.Table(data=[[1, 2, 3]], columns=["a", "b", "c"])
    table3 = wandb.Table(data=[[1, 2, 3]], columns=["a", "b", "c"])
    artifact.add(table0, "table0")
    artifact.add(table1, "nest/table1")
    artifact.add(table2, "nest/nested/table2")
    artifact.add(table3, "nest/nested/nesteded/table3")
    run.log_artifact(artifact)
    run.finish()

    count_node = (
        weave.legacy.weave.ops.project(run.entity, run.project)
        .artifact("test")
        .membershipForAlias("v0")
        .artifactVersion()
        .files()
        .count()
    )
    assert weave.use(count_node) == 4



from weave.legacy.weave import context_state as _context_state

_loading_builtins_token = _context_state.set_loading_built_ins()

# Exporting these
from . import wbgqlquery_op
from . import entity_ops
from . import user_ops
from . import project_ops
from . import run_ops
from . import artifact_type_ops
from . import artifact_collection_ops
from . import artifact_alias_ops
from . import artifact_membership_ops
from . import artifact_version_ops
from . import org_ops
from . import report_ops
from . import run_queue_ops
from . import repo_insight_ops
from . import stream_table_ops

# TODO: Investigate these
from .wbmedia import *
from .table import *
from .trace_tree import *

# from . import wbartifact
# from . import file_wbartifact
# from .. import artifacts_local


# make root ops top level
viewer = user_ops.root_viewer  # type: ignore
project = project_ops.project  # type: ignore
entity = entity_ops.entity  # type: ignore
org = org_ops.org  # type: ignore

_context_state.clear_loading_built_ins(_loading_builtins_token)



import random
import typing

import numpy as np
import pyarrow as pa
from scipy.signal import butter, filtfilt

from weave.legacy.weave import util
from weave.legacy.weave.ops_arrow import ArrowWeaveList

value_fns: list[typing.Any] = [
    lambda steps: steps**2,
    lambda steps: np.cos(steps * 0.0001),
    lambda steps: np.sin(steps * 0.01),
    lambda steps: np.log(steps + 1),
    lambda steps: np.exp(steps * 0.0001),
    lambda steps: np.exp(-steps * 0.0001) * 1000,  # Simulate decreasing loss
    lambda steps: 1 - np.exp(-steps * 0.0001),  # Simulate increasing accuracy
    lambda steps: np.power(steps, -0.5)
    * 1000,  # Simulate decreasing loss with power-law decay
    lambda steps: np.tanh(steps * 0.0001),  # Simulate a metric converging to a value
    lambda steps: np.arctan(
        steps * 0.0001
    ),  # Simulate a metric converging to a value with a different curve
    lambda steps: np.piecewise(
        steps,
        [steps < len(steps) / 2, steps >= len(steps) / 2],
        [lambda steps: steps * 0.001, lambda steps: 1 - np.exp(-steps * 0.0001)],
    ),  # Simulate a two-stage training process
    lambda steps: np.sin(steps * 0.001)
    * np.exp(-steps * 0.0001),  # Sinusoidal oscillations with exponential decay
    lambda steps: (np.cos(steps * 0.001) + 1)
    * 0.5
    * (1 - np.exp(-steps * 0.0001)),  # Oscillations converging to increasing accuracy
    lambda steps: np.log(steps + 1)
    * (
        1 - np.exp(-steps * 0.0001)
    ),  # Logarithmic growth modulated by increasing accuracy
    lambda steps: np.random.random()
    * (
        1 - np.exp(-steps * 0.0001)
    ),  # Random constant value modulated by increasing accuracy
]


def random_metrics_template(n_metrics: int) -> list[int]:
    return [random.randrange(0, len(value_fns)) for _ in range(n_metrics)]


def random_metrics(
    n: int = 100000, n_metrics: int = 100, template: typing.Optional[list[int]] = None
) -> ArrowWeaveList:
    # gpt-4 helped here.

    steps = np.arange(0, n, 1)
    data = {
        "step": steps,
        "string_col": np.random.choice(list("ABCDEFGHIJKLMNOPQRSTUVWXYZ"), n),
    }

    for j in range(n_metrics):
        noise_fraction = random.random()
        if template is None:
            fn = random.choice(value_fns)
        else:
            fn = value_fns[template[j]]
        values = fn(steps)

        # Add different types of noise
        noise_type = random.choice(["uniform", "normal", "triangular"])
        if noise_type == "uniform":
            noise = np.random.uniform(low=-noise_fraction, high=noise_fraction, size=n)
        elif noise_type == "normal":
            noise = np.random.normal(scale=noise_fraction, size=n)
        elif noise_type == "triangular":
            noise = np.random.triangular(
                left=-noise_fraction, mode=0, right=noise_fraction, size=n
            )

        # Apply an optional filter to the noise to simulate more natural variations
        if random.random() < 0.5:
            nyquist = 0.5 * 1
            low = random.uniform(0.01, 0.1) / nyquist
            high = random.uniform(0.1, 0.5) / nyquist
            order = random.randint(1, 5)
            b, a = butter(order, [low, high], btype="band")
            noise = filtfilt(b, a, noise)

        data[f"metric{j}"] = values + noise_fraction * values * noise

    return ArrowWeaveList(pa.table(data))


def random_runs(n_runs: int, n_steps: int, n_metrics: int) -> ArrowWeaveList:
    runs = []
    template = random_metrics_template(n_metrics)
    for i in range(n_runs):
        history = random_metrics(n_steps, n_metrics, template=template)._arrow_data
        hist_array = pa.ListArray.from_arrays([0, len(history)], history)
        config_array = pa.StructArray.from_arrays(
            [[random.choice(["a", "b", "c"])], [random.choice(["x", "y"])]],
            ["a", "b"],
        )
        run_id = util.rand_string_n(8)
        run = pa.StructArray.from_arrays(
            [[run_id], config_array, hist_array], ["id", "config", "history"]
        )
        runs.append(run)
    runs_array = pa.concat_arrays(runs)
    return ArrowWeaveList(runs_array)



import os

import pytest
from anthropic import Anthropic, AsyncAnthropic

import weave

model = "claude-3-haiku-20240307"
# model = "claude-3-opus-20240229"


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
def test_anthropic(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("ANTHROPIC_API_KEY", "DUMMY_API_KEY")
    anthropic_client = Anthropic(
        api_key=api_key,
    )
    message = anthropic_client.messages.create(
        model=model,
        max_tokens=1024,
        messages=[{"role": "user", "content": "Hello, Claude"}],
    )

    all_content = message.content[0]
    exp = "Hello! It's nice to meet you. How can I assist you today?"
    assert all_content.text == exp
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]
    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.id == message.id
    assert output.model == message.model
    assert output.stop_reason == "end_turn"
    assert output.stop_sequence is None
    assert output.content[0].text == exp
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output.model]
    assert model_usage["requests"] == 1
    assert output.usage.output_tokens == model_usage["output_tokens"] == 19
    assert output.usage.input_tokens == model_usage["input_tokens"] == 10


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
def test_anthropic_stream(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("ANTHROPIC_API_KEY", "DUMMY_API_KEY")
    anthropic_client = Anthropic(
        api_key=api_key,
    )
    stream = anthropic_client.messages.create(
        model=model,
        stream=True,
        max_tokens=1024,
        messages=[{"role": "user", "content": "Hello, Claude"}],
    )
    all_content = ""
    for event in stream:
        if event.type == "message_start":
            message = event.message
            input_tokens = event.message.usage.input_tokens
        if event.type == "content_block_delta":
            all_content += event.delta.text
        if event.type == "message_delta":
            output_tokens = event.usage.output_tokens
    exp = "Hello there! How can I assist you today?"
    assert all_content == exp
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.id == message.id
    assert output.model == message.model
    assert output.stop_reason == "end_turn"
    assert output.stop_sequence is None
    assert output.content[0].text == exp
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output.model]
    assert model_usage["requests"] == 1
    assert output.usage.output_tokens == output_tokens == 13
    assert output.usage.input_tokens == input_tokens == 10


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
@pytest.mark.asyncio
async def test_async_anthropic(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    anthropic_client = AsyncAnthropic(
        # This is the default and can be omitted
        api_key=os.environ.get("ANTHROPIC_API_KEY", "DUMMY_API_KEY"),
    )

    message = await anthropic_client.messages.create(
        model=model,
        max_tokens=1024,
        messages=[{"role": "user", "content": "Hello, Claude"}],
    )
    all_content = message.content[0]
    exp = "Hello! It's nice to meet you. How can I assist you today?"
    assert all_content.text == exp
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.id == message.id
    assert output.model == message.model
    assert output.stop_reason == "end_turn"
    assert output.stop_sequence is None
    assert output.content[0].text == exp
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output.model]
    assert model_usage["requests"] == 1
    assert output.usage.output_tokens == model_usage["output_tokens"] == 19
    assert output.usage.input_tokens == model_usage["input_tokens"] == 10


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
@pytest.mark.asyncio
async def test_async_anthropic_stream(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    anthropic_client = AsyncAnthropic(
        # This is the default and can be omitted
        api_key=os.environ.get("ANTHROPIC_API_KEY", "DUMMY_API_KEY"),
    )

    stream = await anthropic_client.messages.create(
        model=model,
        stream=True,
        max_tokens=1024,
        messages=[{"role": "user", "content": "Hello, Claude"}],
    )
    all_content = ""
    async for event in stream:
        if event.type == "message_start":
            message = event.message
            input_tokens = event.message.usage.input_tokens
        if event.type == "content_block_delta":
            all_content += event.delta.text
        if event.type == "message_delta":
            output_tokens = event.usage.output_tokens
    exp = "Hello! It's nice to meet you. How can I assist you today?"
    assert all_content == exp
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.id == message.id
    assert output.model == message.model
    assert output.stop_reason == "end_turn"
    assert output.stop_sequence is None
    assert output.content[0].text == exp
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output.model]
    assert model_usage["requests"] == 1
    assert output.usage.output_tokens == output_tokens == 19
    assert output.usage.input_tokens == input_tokens == 10


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
def test_tools_calling(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("ANTHROPIC_API_KEY", "DUMMY_API_KEY")
    anthropic_client = Anthropic(
        api_key=api_key,
    )
    message = anthropic_client.messages.create(
        model=model,
        max_tokens=1024,
        messages=[
            {
                "role": "user",
                "content": "What's the weather like in San Francisco?",
            }
        ],
        tools=[
            {
                "name": "get_weather",
                "description": "Get the current weather in a given location",
                "input_schema": {
                    "type": "object",
                    "properties": {
                        "location": {
                            "type": "string",
                            "description": "The city and state, e.g. San Francisco, CA",
                        }
                    },
                    "required": ["location"],
                },
            },
        ],
    )
    exp = "San Francisco, CA"
    all_content = message.content[0]
    assert all_content.input["location"] == exp
    assert all_content.type == "tool_use"
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.id == message.id
    assert output.model == message.model
    assert output.stop_reason == "tool_use"
    assert output.stop_sequence is None
    assert output.content[0].input["location"] == exp
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output.model]
    assert model_usage["requests"] == 1
    assert output.usage.output_tokens == model_usage["output_tokens"] == 56
    assert output.usage.input_tokens == model_usage["input_tokens"] == 354


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
def test_anthropic_messages_stream_ctx_manager(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("ANTHROPIC_API_KEY", "DUMMY_API_KEY")
    anthropic_client = Anthropic(
        api_key=api_key,
    )

    all_content = ""
    with anthropic_client.messages.stream(
        max_tokens=1024,
        messages=[
            {
                "role": "user",
                "content": "Say hello there!",
            }
        ],
        model=model,
    ) as stream:
        for event in stream:
            if event.type == "text":
                all_content += event.text

    exp = "Hello there!"
    assert all_content.strip() == exp
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.model == model
    assert output.stop_reason == "end_turn"
    assert output.stop_sequence is None
    assert output.content[0].text.strip() == exp
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output.model]
    assert model_usage["requests"] == 1
    assert output.usage.output_tokens == model_usage["output_tokens"]
    assert output.usage.input_tokens == model_usage["input_tokens"]


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
@pytest.mark.asyncio
async def test_async_anthropic_messages_stream_ctx_manager(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    anthropic_client = AsyncAnthropic(
        api_key=os.environ.get("ANTHROPIC_API_KEY", "DUMMY_API_KEY"),
    )

    all_content = ""
    async with anthropic_client.messages.stream(
        max_tokens=1024,
        messages=[
            {
                "role": "user",
                "content": "Say hello there!",
            }
        ],
        model=model,
    ) as stream:
        async for event in stream:
            if event.type == "text":
                all_content += event.text

    exp = "Hello there!"
    assert all_content.strip() == exp

    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.model == model
    assert output.stop_reason == "end_turn"
    assert output.stop_sequence is None
    assert output.content[0].text.strip() == exp
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output.model]
    assert model_usage["requests"] == 1
    assert output.usage.output_tokens == model_usage["output_tokens"]
    assert output.usage.input_tokens == model_usage["input_tokens"]


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
def test_anthropic_messages_stream_ctx_manager_text(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    api_key = os.environ.get("ANTHROPIC_API_KEY", "DUMMY_API_KEY")
    anthropic_client = Anthropic(
        api_key=api_key,
    )

    all_content = ""
    with anthropic_client.messages.stream(
        max_tokens=1024,
        messages=[
            {
                "role": "user",
                "content": "Say hello there!",
            }
        ],
        model=model,
    ) as stream:
        for text in stream.text_stream:
            all_content += text

    exp = "Hello there!"
    assert all_content.strip() == exp
    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.model == model
    assert output.stop_reason == "end_turn"
    assert output.stop_sequence is None
    assert output.content[0].text.strip() == exp
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output.model]
    assert model_usage["requests"] == 1
    assert output.usage.output_tokens == model_usage["output_tokens"]
    assert output.usage.input_tokens == model_usage["input_tokens"]


@pytest.mark.skip_clickhouse_client
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
@pytest.mark.asyncio
async def test_async_anthropic_messages_stream_ctx_manager_text(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    anthropic_client = AsyncAnthropic(
        api_key=os.environ.get("ANTHROPIC_API_KEY", "DUMMY_API_KEY"),
    )

    all_content = ""
    async with anthropic_client.messages.stream(
        max_tokens=1024,
        messages=[
            {
                "role": "user",
                "content": "Say hello there!",
            }
        ],
        model=model,
    ) as stream:
        async for text in stream.text_stream:
            all_content += text

    exp = "Hello there!"
    assert all_content.strip() == exp

    calls = list(client.calls())
    assert len(calls) == 1
    call = calls[0]

    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.model == model
    assert output.stop_reason == "end_turn"
    assert output.stop_sequence is None
    assert output.content[0].text.strip() == exp
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output.model]
    assert model_usage["requests"] == 1
    assert output.usage.output_tokens == model_usage["output_tokens"]
    assert output.usage.input_tokens == model_usage["input_tokens"]



import time

import numpy as np
import pytest
from PIL import Image

import weave
from weave.legacy.weave import context, execute, gql_json_cache, wandb_api, weave_types
from weave.legacy.weave.wandb_interface.wandb_stream_table import StreamTable


def make_stream_table(*args, **kwargs):
    # Unit test backend does not support async logging
    return StreamTable(*args, **kwargs, _disable_async_file_stream=True)


# Example of end to end integration test
def test_stream_logging(user_by_api_key_in_env):
    st = make_stream_table(
        "test_table",
        project_name="stream-tables",
        entity_name=user_by_api_key_in_env.username,
    )
    for i in range(10):
        st.log({"hello": f"world_{i}", "index": i, "nested": {"a": [i]}})
    st.finish()

    hist_node = (
        weave.legacy.weave.ops.project(user_by_api_key_in_env.username, "stream-tables")
        .run("test_table")
        .history3()
    )

    exp_type = weave_types.TypedDict(
        {"a": weave.types.optional(weave_types.List(weave_types.Int()))}
    )
    nested_type = hist_node.type.value.object_type.property_types["nested"]
    assert exp_type.assign_type(nested_type)
    assert weave.use(hist_node["hello"]).to_pylist_tagged() == [
        f"world_{i}" for i in range(10)
    ]
    assert weave.use(hist_node["index"]).to_pylist_tagged() == [i for i in range(10)]
    assert weave.use(hist_node["nested"]).to_pylist_tagged() == [
        {"a": [i]} for i in range(10)
    ]


def test_bytes_read_from_arrow_reporting(user_by_api_key_in_env):
    st = make_stream_table(
        "test_table",
        project_name="stream-tables",
        entity_name=user_by_api_key_in_env.username,
    )
    for i in range(10):
        st.log({"hello": f"world_{i}", "index": i, "nested": {"a": [i]}})
    st.finish()

    hist_node = (
        weave.legacy.weave.ops.project(user_by_api_key_in_env.username, "stream-tables")
        .run("test_table")
        .history3()
    )

    with execute.top_level_stats() as stats:
        with context.execution_client():
            with gql_json_cache.gql_json_cache_context():
                execute.execute_nodes([hist_node["hello"]])

    # all data is the live set at this point, so it is all counted
    assert stats.summary()["bytes_read_to_arrow"] == 190


def test_stream_logging_image(user_by_api_key_in_env):
    def image():
        imarray = np.random.rand(100, 100, 3) * 255
        return Image.fromarray(imarray.astype("uint8")).convert("RGBA")

    st = make_stream_table(
        "test_table-8",
        project_name="stream-tables",
        entity_name=user_by_api_key_in_env.username,
    )
    for i in range(3):
        st.log({"image": image()})
    st.finish()

    # There is a race case here. For some reason, there is a lag between file uploading
    # and the W&B server being able to properly authenticate the file download request.
    # In lieu of a proper polling/waiting mechanism, we just sleep for a bit. UGLY!
    time.sleep(5)

    hist_node = (
        weave.legacy.weave.ops.project(user_by_api_key_in_env.username, "stream-tables")
        .run("test_table-8")
        .history2()
    )

    images = weave.use(hist_node["image"]).to_pylist_tagged()
    assert len(images) == 3
    assert (np.array(images[0]) != np.array(images[1])).any()
    assert isinstance(images[0], Image.Image)


def test_stream_table_entity_inference(user_by_api_key_in_env):
    st = make_stream_table("stream-tables/test_table-entity-inference")
    for i in range(3):
        st.log({"image": [1, 2, 3]})
    st.finish()

    assert st._entity_name == wandb_api.get_wandb_api_sync().default_entity_name()


def test_multi_writers_sequential(user_by_api_key_in_env):
    st = make_stream_table(
        "test_table",
        project_name="stream-tables",
        entity_name=user_by_api_key_in_env.username,
    )
    indexes = []
    writers = []

    def do_asserts():
        hist_node = (
            weave.legacy.weave.ops.project(user_by_api_key_in_env.username, "stream-tables")
            .run("test_table")
            .history2()
        )
        assert weave.use(hist_node["index"]).to_pylist_tagged() == indexes
        assert weave.use(hist_node["writer"]).to_pylist_tagged() == writers
        assert weave.use(hist_node["_step"]).to_pylist_tagged() == indexes

    for i in range(10):
        indexes.append(i)
        writers.append("a")
        st.log({"index": i, "writer": "a"})

    st.finish()
    do_asserts()

    st = make_stream_table(
        "test_table",
        project_name="stream-tables",
        entity_name=user_by_api_key_in_env.username,
    )
    for i in range(10):
        indexes.append(10 + i)
        writers.append("b")
        st.log({"index": 10 + i, "writer": "b"})
    st.finish()

    do_asserts()


@pytest.mark.skip(reason="Multi-writer not yet supported in local container")
def test_multi_writers_parallel(user_by_api_key_in_env):
    entity_name = user_by_api_key_in_env.username
    table_name = "test_table_" + str(int(time.time()))
    st_1 = make_stream_table(
        table_name,
        project_name="stream-tables",
        entity_name=entity_name,
    )
    st_2 = make_stream_table(
        table_name,
        project_name="stream-tables",
        entity_name=entity_name,
    )

    indexes = []
    writers = []

    for i in range(100):
        st_1.log({"index": i * 2, "writer": "a"})
        st_2.log({"index": i * 2 + 1, "writer": "b"})
        indexes.append(i * 2)
        indexes.append(i * 2 + 1)
        writers.append("a")
        writers.append("b")
        time.sleep(0.025)
    st_1.finish()
    st_2.finish()

    hist_node = (
        weave.legacy.weave.ops.project(entity_name, "stream-tables")
        .run(table_name)
        .history2()
    )
    assert weave.use(hist_node["index"]).to_pylist_raw() == indexes
    assert weave.use(hist_node["writer"]).to_pylist_raw() == writers
    assert weave.use(hist_node["_step"]).to_pylist_raw() == [i for i in range(20)]


def test_stream_unauthed(user_by_api_key_in_env):
    with pytest.raises(weave.errors.WeaveWandbAuthenticationException):
        st = make_stream_table(
            "test_table",
            project_name="stream-tables",
            entity_name="NONEXISTENT_USER",
        )


def test_stream_authed(user_by_api_key_in_env):
    st = make_stream_table(
        "test_table",
        project_name="stream-tables",
        entity_name=user_by_api_key_in_env.username,
    )
    st.log({"hello": "world"})
    st.finish()

    a = weave.use(st.rows()["hello"]).to_pylist_tagged()
    assert a == ["world"]


def test_stream_templates(user_by_api_key_in_env):
    st = make_stream_table(
        "test_table",
        project_name="stream-tables",
        entity_name=user_by_api_key_in_env.username,
    )
    st.log({"hello": "world"})
    st.finish()

    a = weave.use(st.rows().get_board_templates_for_node())
    assert len(a) > 0



import typing

from pydantic import BaseModel

from . import refs_internal as ri

A = typing.TypeVar("A")
B = typing.TypeVar("B")

weave_prefix = ri.WEAVE_SCHEME + ":///"
weave_internal_prefix = ri.WEAVE_INTERNAL_SCHEME + ":///"


class InvalidExternalRef(ValueError):
    pass


class InvalidInternalRef(ValueError):
    pass


def universal_ext_to_int_ref_converter(
    obj: A, convert_ext_to_int_project_id: typing.Callable[[str], str]
) -> A:
    """Takes any object and recursively replaces all external references with
    internal references. The external references are expected to be in the
    format of `weave:///entity/project/...` and the internal references are
    expected to be in the format of `weave-trace-internal:///project_id/...`.

    Args:
        obj: The object to convert.
        convert_ext_to_int_project_id: A function that takes an external
            project ID and returns the internal project ID.

    Returns:
        The object with all external references replaced with internal
        references.
    """
    ext_to_int_project_cache: typing.Dict[str, str] = {}

    def replace_ref(ref_str: str) -> str:
        if not ref_str.startswith(weave_prefix):
            raise ValueError(f"Invalid URI: {ref_str}")
        rest = ref_str[len(weave_prefix) :]
        parts = rest.split("/", 2)
        if len(parts) != 3:
            raise InvalidExternalRef(f"Invalid URI: {ref_str}")
        entity, project, tail = parts
        project_key = f"{entity}/{project}"
        if project_key not in ext_to_int_project_cache:
            ext_to_int_project_cache[project_key] = convert_ext_to_int_project_id(
                project_key
            )
        internal_project_id = ext_to_int_project_cache[project_key]
        return f"{ri.WEAVE_INTERNAL_SCHEME}:///{internal_project_id}/{tail}"

    def mapper(obj: B) -> B:
        if isinstance(obj, str):
            if obj.startswith(weave_prefix):
                return typing.cast(B, replace_ref(obj))
            elif obj.startswith(weave_internal_prefix):
                # It is important to raise here as this would be the result of
                # an external client attempting to write internal refs directly.
                # We want to maintain full control over the internal refs.
                raise InvalidExternalRef("Encountered unexpected ref format.")
        return obj

    return _map_values(obj, mapper)


C = typing.TypeVar("C")
D = typing.TypeVar("D")


def universal_int_to_ext_ref_converter(
    obj: C,
    convert_int_to_ext_project_id: typing.Callable[[str], typing.Optional[str]],
) -> C:
    """Takes any object and recursively replaces all internal references with
    external references. The internal references are expected to be in the
    format of `weave-trace-internal:///project_id/...` and the external references are
    expected to be in the format of `weave:///entity/project/...`.

    Args:
        obj: The object to convert.
        convert_int_to_ext_project_id: A function that takes an internal
            project ID and returns the external project ID.

    Returns:
        The object with all internal references replaced with external
        references.
    """
    int_to_ext_project_cache: dict[str, typing.Optional[str]] = {}

    def replace_ref(ref_str: str) -> str:
        if not ref_str.startswith(weave_internal_prefix):
            raise ValueError(f"Invalid URI: {ref_str}")
        rest = ref_str[len(weave_internal_prefix) :]
        parts = rest.split("/", 1)
        if len(parts) != 2:
            raise InvalidInternalRef(f"Invalid URI: {ref_str}")
        project_id, tail = parts
        if project_id not in int_to_ext_project_cache:
            int_to_ext_project_cache[project_id] = convert_int_to_ext_project_id(
                project_id
            )
        external_project_id = int_to_ext_project_cache[project_id]
        if not external_project_id:
            return f"{ri.WEAVE_PRIVATE_SCHEME}://///{tail}"
        return f"{ri.WEAVE_SCHEME}:///{external_project_id}/{tail}"

    def mapper(obj: D) -> D:
        if isinstance(obj, str):
            if obj.startswith(weave_internal_prefix):
                return typing.cast(D, replace_ref(obj))
            elif obj.startswith(weave_prefix):
                # It is important to raise here as this would be the result of
                # incorrectly storing an external ref at the database layer,
                # rather than an internal ref. There is a possibility in the
                # future that a programming error leads to this situation, in
                # which case reading this object would consistently fail. We
                # might want to instead return a private ref in this case.
                raise InvalidInternalRef("Encountered unexpected ref format.")
        return obj

    return _map_values(obj, mapper)


E = typing.TypeVar("E")
F = typing.TypeVar("F")


def _map_values(obj: E, func: typing.Callable[[E], E]) -> E:
    if isinstance(obj, BaseModel):
        # `by_alias` is required since we have Mongo-style properties in the
        # query models that are aliased to conform to start with `$`. Without
        # this, the model_dump will use the internal property names which are
        # not valid for the `model_validate` step.
        orig = obj.model_dump(by_alias=True)
        new = _map_values(orig, func)
        return obj.model_validate(new)
    if isinstance(obj, dict):
        return typing.cast(E, {k: _map_values(v, func) for k, v in obj.items()})
    if isinstance(obj, list):
        return typing.cast(E, [_map_values(v, func) for v in obj])
    if isinstance(obj, tuple):
        return typing.cast(E, tuple(_map_values(v, func) for v in obj))
    if isinstance(obj, set):
        return typing.cast(E, {_map_values(v, func) for v in obj})
    return func(obj)



import pytest

import weave
import weave.legacy.weave

ops = weave.legacy.weave.registry_mem.memory_registry.list_ops()


def output_type_dict_is_const_function_node(output_type_dict):
    return (
        "nodeType" in output_type_dict
        and output_type_dict["nodeType"] == "const"
        and "type" in output_type_dict["type"]
        and output_type_dict["type"]["type"] == "function"
    )


def const_funtion_node_val(node_dict):
    return node_dict["val"]


def assert_node_dict_free_of_get_ops(node_dict):
    if node_dict["nodeType"] == "output":
        assert node_dict["fromOp"]["name"] != "get"
        for input_node in node_dict["fromOp"]["inputs"].values():
            assert_node_dict_free_of_get_ops(input_node)


def assert_valid_output_type(op_dict):
    output_type_dict = op_dict["output_type"]
    if not output_type_dict_is_const_function_node(output_type_dict):
        return
    assert_node_dict_free_of_get_ops(const_funtion_node_val(output_type_dict))


@pytest.mark.parametrize("op_name, op", [(op.name, op) for op in ops])
def test_explicit_experiment_construction(op_name, op):
    # Just make sure that this is successful
    try:
        op_as_dict = op.to_dict()
    except weave.errors.WeaveSerializeError:
        # Don't check these for now, it indicates the op had a callable output type.
        # We are not currently sending those to WeaveJS
        return
    assert op_as_dict is not None

    print(op_as_dict)
    assert_valid_output_type(op_as_dict)



from weave.legacy.weave import api as weave
from weave.legacy.weave import box, ops
from weave.legacy.weave.ops_primitives import number
from weave.legacy.weave.ops_primitives.string import *
from weave.legacy.weave.weave_internal import make_const_node


def test_number_ops():
    nine = make_const_node(weave.types.Number(), 9)
    assert weave.use(nine + 3) == 12
    assert weave.use(nine - 3) == 6
    assert weave.use(nine * 3) == 27
    assert weave.use(nine / 2) == 4.5
    assert weave.use(nine % 3) == 0
    assert weave.use(nine**2) == 81

    assert weave.use(nine == 8) == False
    assert weave.use(nine == 9) == True
    assert weave.use(nine != 8) == True
    assert weave.use(nine != 9) == False
    assert weave.use(nine > 8) == True
    assert weave.use(nine > 9) == False
    assert weave.use(nine < 10) == True
    assert weave.use(nine < 9) == False
    assert weave.use(nine >= 8) == True
    assert weave.use(nine >= 9) == True
    assert weave.use(nine <= 10) == True
    assert weave.use(nine <= 9) == True
    assert weave.use(-nine) == -9
    assert weave.use(nine.pow(2)) == 81
    assert weave.use(nine.toString()) == "9"

    pi = make_const_node(weave.types.Number(), 3.14)

    assert weave.use(pi.floor()) == 3
    assert weave.use(pi.ceil()) == 4

    one = make_const_node(weave.types.Number(), 1)
    two = make_const_node(weave.types.Number(), 2)
    three = make_const_node(weave.types.Number(), 3)
    list_of_nums = ops.make_list(x1=one, x2=two, x3=three)

    assert weave.use(number.numbers_sum(list_of_nums)) == 6
    assert weave.use(number.numbers_avg(list_of_nums)) == 2
    assert weave.use(number.numbers_min(list_of_nums)) == 1
    assert weave.use(number.numbers_max(list_of_nums)) == 3


def test_string_ops():
    foo = make_const_node(weave.types.String(), "Foo")

    assert weave.use(string(foo)) == "Foo"
    assert weave.use(lastLetter(foo)) == "o"
    assert weave.use(foo == "Foo") == True
    assert weave.use(foo == "Baz") == False
    assert weave.use(foo != "Baz") == True
    assert weave.use(foo != "Foo") == False
    assert weave.use(foo + "Bar") == "FooBar"
    assert weave.use(foo.len()) == 3
    assert weave.use(foo.append("Bar")) == "FooBar"
    assert weave.use(foo.prepend("Bar")) == "BarFoo"
    assert weave.use(foo.lower()) == "foo"
    assert weave.use(foo.upper()) == "FOO"
    assert weave.use(foo.slice(0, 2)) == "Fo"
    assert weave.use(foo.slice(1, 3)) == "oo"
    assert weave.use(foo.replace("oo", "xx")) == "Fxx"

    foobar = make_const_node(weave.types.String(), "Foo bar baz")
    assert weave.use(foobar.split(" ")) == ["Foo", "bar", "baz"]
    assert weave.use(foobar.partition("bar")) == ["Foo ", "bar", " baz"]
    assert weave.use(foobar.startsWith("Foo")) == True
    assert weave.use(foobar.startsWith("bar")) == False
    assert weave.use(foobar.endsWith("baz")) == True
    assert weave.use(foobar.endsWith("bar")) == False

    num_string = make_const_node(weave.types.String(), "123")
    alpha_string = make_const_node(weave.types.String(), "ABC")

    assert weave.use(num_string.isAlpha()) == False
    assert weave.use(num_string.isNumeric()) == True
    assert weave.use(num_string.isAlnum()) == True
    assert weave.use(alpha_string.isAlpha()) == True
    assert weave.use(alpha_string.isNumeric()) == False
    assert weave.use(alpha_string.isAlnum()) == True

    foo_space = make_const_node(weave.types.String(), "  Foo  ")
    assert weave.use(foo_space.strip()) == "Foo"
    assert weave.use(foo_space.lStrip()) == "Foo  "
    assert weave.use(foo_space.rStrip()) == "  Foo"

    assert weave.use(num_string.toNumber()) == 123
    assert weave.use(alpha_string.toNumber()) == None

    # assert weave.use(foo in foobar) == True # Broken
    # assert weave.use(foobar in foo) == False # Broken


def test_null_consuming_numbers_ops():
    data = [box.box(1), box.box(None), box.box(2)]

    assert weave.use(number.numbers_sum(data)) == 3
    assert weave.use(number.numbers_avg(data)) == 1.5
    assert weave.use(number.numbers_min(data)) == 1
    assert weave.use(number.numbers_max(data)) == 2
    assert number.numbers_max(data).type == weave.types.optional(weave.types.Int())
    assert number.numbers_avg(data).type == weave.types.optional(weave.types.Number())

    data = [box.box(None), box.box(None), box.box(None)]
    assert weave.use(number.numbers_sum(data)) == None
    assert weave.use(number.numbers_avg(data)) == None
    assert weave.use(number.numbers_min(data)) == None
    assert weave.use(number.numbers_max(data)) == None
    assert number.numbers_max(data).type == weave.types.NoneType()

    data = []
    assert weave.use(number.numbers_sum(data)) == None
    assert weave.use(number.numbers_avg(data)) == None
    assert weave.use(number.numbers_min(data)) == None
    assert weave.use(number.numbers_max(data)) == None
    assert number.numbers_max(data).type == weave.types.NoneType()


def test_null_vararg_ops():
    assert weave.use(ops.dict_(index=None, id=None)) == {"index": None, "id": None}
    assert weave.use(ops.make_list(index=None, id=None)) == [None, None]



import typing

from weave.legacy.weave import weave_types as types
from weave.legacy.weave.api import op, weave_class
from weave.legacy.weave import dispatch
from weave.legacy.weave.ops_primitives.dict import dict_


@weave_class(weave_type=types.Boolean)
class Boolean:
    @op(name="boolean-equal")
    def bool_equals(lhs: typing.Optional[bool], rhs: typing.Optional[bool]) -> bool:  # type: ignore
        return lhs == rhs

    @op(name="boolean-notEqual")
    def bool_notEquals(lhs: typing.Optional[bool], rhs: typing.Optional[bool]) -> bool:  # type: ignore
        return lhs != rhs

    @op(name="and")
    def bool_and(lhs: bool, rhs: typing.Optional[bool]) -> bool:  # type: ignore
        return lhs and bool(rhs)

    @op(name="or")
    def bool_or(lhs: bool, rhs: typing.Optional[bool]) -> bool:  # type: ignore
        return lhs or bool(rhs)

    @op(name="boolean-not")
    def bool_not(bool: bool) -> bool:  # type: ignore
        return not bool


types.Boolean.instance_class = Boolean


def none_coalesce_output_type(input_types):
    if types.NoneType().assign_type(input_types["lhs"]):
        return input_types["rhs"]
    return types.union(input_types["lhs"], input_types["rhs"])


@op(name="none-coalesce", output_type=none_coalesce_output_type)
def none_coalesce(lhs: typing.Any, rhs: typing.Any):
    # TODO: This logic is really complicated in Weave0.
    if isinstance(lhs, list) and isinstance(rhs, list):
        return [l or r for l, r in zip(lhs, rhs)]
    return lhs or rhs


@op(
    name="cond",
    output_type=lambda input_type: types.optional(
        types.union(*input_type["results"].property_types.values())
    ),
)
def cond(cases: dict[str, bool], results: dict[str, typing.Any]):
    """Return first result.values()[i] for which case.values()[i] is True.

    Can be used where to achieve if/then/else, or SQL CASE-like behavior.

    Arguments are dictionaries, not lists, but keys are ignored entirely.
    This is important for vectorization. Lists are stored row-oriented in arrow,
    but we actually want columns for each case and result. Using dictionaries
    achieve this, since we map them to arrow structs.
    """
    for c, r in zip(cases.values(), results.values()):
        if c:
            return r
    return None


class Case(typing.TypedDict):
    # These are both actually Nodes, but this gets type-checking to pass for now.
    when: bool
    then: typing.Any


def case(cases: list[Case]) -> dispatch.RuntimeOutputNode:
    """Helper for writing cond expressions."""
    sep_cases = {"%s" % i: cases[i]["when"] for i in range(len(cases))}
    sep_results = {"%s" % i: cases[i]["then"] for i in range(len(cases))}
    return cond(dict_(**sep_cases), dict_(**sep_results))



import dataclasses
import typing

import weave
from weave.legacy.weave import graph, panel
from weave.legacy.weave.panels.panel_group import PanelBankSectionConfig

RenderType = typing.TypeVar("RenderType")


@weave.type()
class SectionsConfig(typing.Generic[RenderType]):
    section: weave.Node[typing.Optional[typing.Any]] = dataclasses.field(
        default_factory=lambda: weave.legacy.weave.graph.VoidNode()
    )
    panel: RenderType = dataclasses.field(default_factory=lambda: graph.VoidNode())  # type: ignore


@weave.type()
class Sections(panel.Panel):
    id = "Sections"
    config: typing.Optional[SectionsConfig] = dataclasses.field(
        default_factory=lambda: None
    )



from PIL import Image

from weave.legacy.weave import api as weave
from weave.legacy.weave import context_state
from weave.legacy.weave.ops_primitives import geom as media_user


def test_im_with_metadata():
    base_im = Image.linear_gradient("L").resize((32, 32))
    ims = [
        media_user.ImageWithBoxes(
            base_im,
            [
                media_user.BoundingBox2D(
                    media_user.Point2D(1, 2), media_user.Size2D(5, 9)
                )
            ],
        ),
        media_user.ImageWithBoxes(
            base_im.rotate(4),
            [
                media_user.BoundingBox2D(
                    media_user.Point2D(5, 9), media_user.Size2D(1, 2)
                )
            ],
        ),
    ]
    ims_saved = weave.save(ims)
    assert (
        weave.use(
            ims_saved.map(
                lambda im: im.get_boxes().map(lambda box: box.center().get_x()).sum()
            ).sum()
        )
        == 9.0
    )



from weave.trace.client_context import weave_client as weave_client_context
from weave.trace.patcher import Patcher
from weave.trace.weave_client import Call

TRANSFORM_EMBEDDINGS = False
ALLOWED_ROOT_EVENT_TYPES = ("query",)

import_failed = False

try:
    from llama_index.core.callbacks.base_handler import BaseCallbackHandler
    from llama_index.core.callbacks.schema import CBEventType, EventPayload
except ImportError:
    # This occurs if llama_index is not installed.
    import_failed = True
except Exception:
    # This occurs if llama_index is installed but there is an error in the import or some other error occured in the interaction between packages.
    import_failed = True
    print(
        "Failed to autopatch llama_index. If you are tracing Llama calls, please upgrade llama_index to be version>=0.10.35"
    )


from typing import Any, Dict, List, Optional, Tuple

if not import_failed:

    class WeaveCallbackHandler(BaseCallbackHandler):
        """Base callback handler that can be used to track event starts and ends."""

        def __init__(
            self,
            event_starts_to_ignore: Optional[List[CBEventType]] = None,
            event_ends_to_ignore: Optional[List[CBEventType]] = None,
        ) -> None:
            # Map from event id to call object - this is used for bookkeeping
            # and eventually closing the call.
            self._call_map: Dict[str, Call] = {}

            # Everything below here is just boilerplate for inheriting from
            # BaseCallbackHandler.
            event_starts_to_ignore = (
                event_starts_to_ignore if event_starts_to_ignore else []
            )
            event_ends_to_ignore = event_ends_to_ignore if event_ends_to_ignore else []
            super().__init__(
                event_starts_to_ignore=event_starts_to_ignore,
                event_ends_to_ignore=event_ends_to_ignore,
            )

        def on_event_start(
            self,
            event_type: CBEventType,
            payload: Optional[Dict[EventPayload, Any]] = None,
            event_id: str = "",
            parent_id: str = "",
            **kwargs: Any,
        ) -> str:
            """Run when an event starts and return id of event."""
            # Get a handle to the internal graph client.
            gc = weave_client_context.require_weave_client()

            # Check to see if the event is an exception.
            if event_type == CBEventType.EXCEPTION:
                # If the event is an exception, and we are actively tracking the corresponding
                # call, finish the call with the exception.
                if event_id in self._call_map:
                    # Pop the call from the call map.
                    call = self._call_map.pop(event_id)

                    # Get the exception from the payload if it exists, otherwise use a default message.
                    if payload:
                        exception = payload.get("EXCEPTION")
                    else:
                        exception = "Unknown exception occurred."

                    # Finish the call with the exception.
                    gc.finish_call(call, None, exception=exception)
            else:
                # Check if the event is a valid root event or child event.
                # Here, we only allow a subset of event types as the root since not
                # all event types are meaningful in the weave context.
                is_valid_root = (
                    parent_id == "root" and event_type in ALLOWED_ROOT_EVENT_TYPES
                )
                # Since we don't track all calls, it is possible that the parent_id
                # is not in the call map. In this case, we don't want to track the child event.
                is_valid_child = parent_id != "root" and parent_id in self._call_map

                # If the event is valid, create a call and add it to the call map.
                if is_valid_root or is_valid_child:
                    # Create a call object.
                    call = gc.create_call(
                        "llama_index." + event_type.name.lower(),
                        process_payload(payload),
                        self._call_map.get(parent_id),
                    )

                    # Add the call to the call map.
                    self._call_map[event_id] = call

            # Return the event id (this is just part of interface requirements)
            return event_id

        def on_event_end(
            self,
            event_type: CBEventType,
            payload: Optional[Dict[EventPayload, Any]] = None,
            event_id: str = "",
            **kwargs: Any,
        ) -> None:
            """Run when an event ends."""
            # Get a handle to the internal graph client.
            gc = weave_client_context.require_weave_client()

            # If the event is in the call map, finish the call.
            if event_id in self._call_map:
                # Finish the call.
                call = self._call_map.pop(event_id)
                gc.finish_call(call, process_payload(payload))

        def start_trace(self, trace_id: Optional[str] = None) -> None:
            """Run when an overall trace is launched."""
            # Not implemented - required by interface.
            pass

        def end_trace(
            self,
            trace_id: Optional[str] = None,
            trace_map: Optional[Dict[str, List[str]]] = None,
        ) -> None:
            """Run when an overall trace is exited."""
            # Not implemented - required by interface.
            pass

    def process_payload(
        payload: Optional[Dict[EventPayload, Any]] = None,
    ) -> Optional[Dict[EventPayload, Any]]:
        if payload is None:
            return None
        res = {}
        for k, v in payload.items():
            if TRANSFORM_EMBEDDINGS and k == EventPayload.EMBEDDINGS:
                shape = get_embedding_shape(v)
                res[k] = f"Embedding with shape {shape}"
            else:
                res[k] = v
        return res

    def get_embedding_shape(embedding: List) -> Tuple:
        """Get the shape of an embedding."""
        res = []
        while isinstance(embedding, list):
            res.append(len(embedding))
            embedding = embedding[0]
        return tuple(res)

else:

    class WeaveCallbackHandler:  # type: ignore
        pass


class LLamaIndexPatcher(Patcher):
    def __init__(self) -> None:
        pass

    def attempt_patch(self) -> bool:
        if import_failed:
            return False
        try:
            import llama_index.core

            self._original_handler = llama_index.core.global_handler

            llama_index.core.global_handler = WeaveCallbackHandler()
            return True
        except Exception:
            return False

    def undo_patch(self) -> bool:
        if not hasattr(self, "_original_handler"):
            return False
        try:
            import llama_index.core

            llama_index.core.global_handler = self._original_handler
            return True
        except Exception:
            return False


llamaindex_patcher = LLamaIndexPatcher()



import sqlparse

from weave.trace_server import trace_server_interface as tsi
from weave.trace_server.calls_query_builder import CallsQuery, HardCodedFilter
from weave.trace_server.interface import query as tsi_query
from weave.trace_server.orm import ParamBuilder


def test_query_baseline() -> None:
    cq = CallsQuery(project_id="project")
    cq.add_field("id")
    assert_sql(
        cq,
        """
        SELECT calls_merged.id AS id
        FROM calls_merged
        WHERE project_id = {pb_0:String}
        GROUP BY (project_id,id)
        HAVING (
            ((
                any(calls_merged.deleted_at) IS NULL
            ))
            AND
            ((
               NOT ((
                  any(calls_merged.started_at) IS NULL
               ))
            ))
        )
        """,
        {"pb_0": "project"},
    )


def test_query_light_column() -> None:
    cq = CallsQuery(project_id="project")
    cq.add_field("id")
    cq.add_field("started_at")
    assert_sql(
        cq,
        """
        SELECT
            calls_merged.id AS id,
            any(calls_merged.started_at) AS started_at
        FROM calls_merged
        WHERE project_id = {pb_0:String}
        GROUP BY (project_id,id)
        HAVING (
            ((
                any(calls_merged.deleted_at) IS NULL
            ))
            AND
            ((
               NOT ((
                  any(calls_merged.started_at) IS NULL
               ))
            ))
        )
        """,
        {"pb_0": "project"},
    )


def test_query_heavy_column() -> None:
    cq = CallsQuery(project_id="project")
    cq.add_field("id")
    cq.add_field("inputs")
    assert_sql(
        cq,
        """
        SELECT
            calls_merged.id AS id,
            any(calls_merged.inputs_dump) AS inputs_dump
        FROM calls_merged
        WHERE project_id = {pb_0:String}
        GROUP BY (project_id,id)
        HAVING (
            ((
                any(calls_merged.deleted_at) IS NULL
            ))
            AND
            ((
               NOT ((
                  any(calls_merged.started_at) IS NULL
               ))
            ))
        )
        """,
        {"pb_0": "project"},
    )


def test_query_heavy_column_simple_filter() -> None:
    cq = CallsQuery(project_id="project")
    cq.add_field("id")
    cq.add_field("inputs")
    cq.set_hardcoded_filter(
        HardCodedFilter(
            filter=tsi.CallsFilter(
                op_names=["a", "b"],
            )
        )
    )
    assert_sql(
        cq,
        """
        WITH filtered_calls AS (
            SELECT
                calls_merged.id AS id
            FROM calls_merged
            WHERE project_id = {pb_1:String}
            GROUP BY (project_id,id)
            HAVING (
                ((any(calls_merged.deleted_at) IS NULL))
                AND ((NOT ((any(calls_merged.started_at) IS NULL))))
            AND
                (any(calls_merged.op_name) IN {pb_0:Array(String)}))
        )
        SELECT
            calls_merged.id AS id,
            any(calls_merged.inputs_dump) AS inputs_dump
        FROM calls_merged
        WHERE
            project_id = {pb_2:String}
        AND
            (id IN filtered_calls)
        GROUP BY (project_id,id)
        """,
        {"pb_0": ["a", "b"], "pb_1": "project", "pb_2": "project"},
    )


def test_query_heavy_column_simple_filter_with_order() -> None:
    cq = CallsQuery(project_id="project")
    cq.add_field("id")
    cq.add_field("inputs")
    cq.add_order("started_at", "desc")
    cq.set_hardcoded_filter(
        HardCodedFilter(
            filter=tsi.CallsFilter(
                op_names=["a", "b"],
            )
        )
    )
    assert_sql(
        cq,
        """
        WITH filtered_calls AS (
            SELECT
                calls_merged.id AS id
            FROM calls_merged
            WHERE project_id = {pb_1:String}
            GROUP BY (project_id,id)
            HAVING (
                ((any(calls_merged.deleted_at) IS NULL))
                AND ((NOT ((any(calls_merged.started_at) IS NULL))))
            AND
                (any(calls_merged.op_name) IN {pb_0:Array(String)}))
        )
        SELECT
            calls_merged.id AS id,
            any(calls_merged.inputs_dump) AS inputs_dump
        FROM calls_merged
        WHERE
            project_id = {pb_2:String}
        AND
            (id IN filtered_calls)
        GROUP BY (project_id,id)
        ORDER BY any(calls_merged.started_at) DESC
        """,
        {"pb_0": ["a", "b"], "pb_1": "project", "pb_2": "project"},
    )


def test_query_heavy_column_simple_filter_with_order_and_limit() -> None:
    cq = CallsQuery(project_id="project")
    cq.add_field("id")
    cq.add_field("inputs")
    cq.add_order("started_at", "desc")
    cq.set_limit(10)
    cq.set_hardcoded_filter(
        HardCodedFilter(
            filter=tsi.CallsFilter(
                op_names=["a", "b"],
            )
        )
    )
    assert_sql(
        cq,
        """
        WITH filtered_calls AS (
            SELECT
                calls_merged.id AS id
            FROM calls_merged
            WHERE project_id = {pb_1:String}
            GROUP BY (project_id,id)
            HAVING (
                ((any(calls_merged.deleted_at) IS NULL))
            AND
                ((NOT ((any(calls_merged.started_at) IS NULL))))
            AND
                (any(calls_merged.op_name) IN {pb_0:Array(String)})
            )
            ORDER BY any(calls_merged.started_at) DESC
            LIMIT 10
        )
        SELECT
            calls_merged.id AS id,
            any(calls_merged.inputs_dump) AS inputs_dump
        FROM calls_merged
        WHERE
            project_id = {pb_2:String}
        AND
            (id IN filtered_calls)
        GROUP BY (project_id,id)
        ORDER BY any(calls_merged.started_at) DESC
        """,
        {"pb_0": ["a", "b"], "pb_1": "project", "pb_2": "project"},
    )


def test_query_heavy_column_simple_filter_with_order_and_limit_and_mixed_query_conditions() -> (
    None
):
    cq = CallsQuery(project_id="project")
    cq.add_field("id")
    cq.add_field("inputs")
    cq.add_order("started_at", "desc")
    cq.set_limit(10)
    cq.set_hardcoded_filter(
        HardCodedFilter(
            filter=tsi.CallsFilter(
                op_names=["a", "b"],
            )
        )
    )
    cq.add_condition(
        tsi_query.AndOperation.model_validate(
            {
                "$and": [
                    {
                        "$eq": [
                            {"$getField": "inputs.param.val"},
                            {"$literal": "hello"},
                        ]
                    },  # <-- heavy condition
                    {
                        "$eq": [{"$getField": "wb_user_id"}, {"$literal": "my_user_id"}]
                    },  # <-- light condition
                ]
            }
        )
    )
    assert_sql(
        cq,
        """
        WITH filtered_calls AS (
            SELECT
                calls_merged.id AS id
            FROM calls_merged
            WHERE project_id = {pb_2:String}
            GROUP BY (project_id,id)
            HAVING (
                ((any(calls_merged.wb_user_id) = {pb_0:String}))
            AND
                ((any(calls_merged.deleted_at) IS NULL))
            AND
                ((NOT ((any(calls_merged.started_at) IS NULL))))
            AND
                (any(calls_merged.op_name) IN {pb_1:Array(String)})
            )
        )
        SELECT
            calls_merged.id AS id,
            any(calls_merged.inputs_dump) AS inputs_dump
        FROM calls_merged
        WHERE
            project_id = {pb_5:String}
        AND
            (id IN filtered_calls)
        GROUP BY (project_id,id)
        HAVING (
            JSON_VALUE(any(calls_merged.inputs_dump), {pb_3:String}) = {pb_4:String}
        )
        ORDER BY any(calls_merged.started_at) DESC
        LIMIT 10
        """,
        {
            "pb_0": "my_user_id",
            "pb_1": ["a", "b"],
            "pb_2": "project",
            "pb_3": '$."param"."val"',
            "pb_4": "hello",
            "pb_5": "project",
        },
    )


def assert_sql(cq: CallsQuery, exp_query, exp_params):
    pb = ParamBuilder("pb")
    query = cq.as_sql(pb)
    params = pb.get_params()

    assert exp_params == params

    exp_formatted = sqlparse.format(exp_query, reindent=True)
    found_formatted = sqlparse.format(query, reindent=True)

    assert exp_formatted == found_formatted


def test_query_light_column_with_costs() -> None:
    cq = CallsQuery(
        project_id="UHJvamVjdEludGVybmFsSWQ6Mzk1NDg2Mjc=", include_costs=True
    )
    cq.add_field("id")
    cq.add_field("started_at")
    cq.set_hardcoded_filter(
        HardCodedFilter(
            filter=tsi.CallsFilter(
                op_names=["a", "b"],
            )
        )
    )
    assert_sql(
        cq,
        """
        WITH
            filtered_calls AS (
                SELECT calls_merged.id AS id
                FROM calls_merged
                WHERE project_id = {pb_1:String}
                GROUP BY (project_id, id)
                HAVING (((any(calls_merged.deleted_at) IS NULL))
                AND ((NOT ((any(calls_merged.started_at) IS NULL))))
                AND (any(calls_merged.op_name) IN {pb_0:Array(String)}))),
            all_calls AS (
                SELECT
                    calls_merged.id AS id,
                    any(calls_merged.started_at) AS started_at
                FROM calls_merged
                WHERE project_id = {pb_2:String}
                    AND (id IN filtered_calls)
                GROUP BY (project_id, id)),
            -- From the all_calls we get the usage data for LLMs
            llm_usage AS (
                SELECT
                    *,
                    ifNull(JSONExtractRaw(summary_dump, 'usage'), '{}') AS usage_raw,
                    arrayJoin(
                        if(usage_raw != '',
                        JSONExtractKeysAndValuesRaw(usage_raw),
                        [('weave_dummy_llm_id', '{"requests": 0, "prompt_tokens": 0, "completion_tokens": 0, "total_tokens": 0}')])
                    ) AS kv,
                    kv.1 AS llm_id,
                    JSONExtractInt(kv.2, 'requests') AS requests,
                    if(JSONHas(kv.2, 'prompt_tokens'), JSONExtractInt(kv.2, 'prompt_tokens'), JSONExtractInt(kv.2, 'input_tokens')) AS prompt_tokens,
                    if(JSONHas(kv.2, 'completion_tokens'), JSONExtractInt(kv.2, 'completion_tokens'), JSONExtractInt(kv.2, 'output_tokens')) AS completion_tokens,
                    JSONExtractInt(kv.2, 'total_tokens') AS total_tokens
                FROM all_calls),
            -- based on the llm_ids in the usage data we get all the prices and rank them according to specificity and effective date
            ranked_prices AS (
                SELECT
                    *,
                    llm_token_prices.id,
                    llm_token_prices.pricing_level,
                    llm_token_prices.pricing_level_id,
                    llm_token_prices.provider_id,
                    llm_token_prices.llm_id,
                    llm_token_prices.effective_date,
                    llm_token_prices.prompt_token_cost,
                    llm_token_prices.completion_token_cost,
                    llm_token_prices.prompt_token_cost_unit,
                    llm_token_prices.completion_token_cost_unit,
                    llm_token_prices.created_by,
                    llm_token_prices.created_at,
                    ROW_NUMBER() OVER (
                        PARTITION BY llm_usage.id, llm_usage.llm_id
                        ORDER BY
                            CASE
                                -- Order by effective_date
                                WHEN llm_usage.started_at >= llm_token_prices.effective_date THEN 1
                                ELSE 2
                            END,
                            CASE
                                -- Order by pricing level then by effective_date
                                -- WHEN llm_token_prices.pricing_level = 'org' AND llm_token_prices.pricing_level_id = ORG_PARAM THEN 1
                                WHEN llm_token_prices.pricing_level = 'project' AND llm_token_prices.pricing_level_id = 'UHJvamVjdEludGVybmFsSWQ6Mzk1NDg2Mjc=' THEN 2
                                WHEN llm_token_prices.pricing_level = 'default' AND llm_token_prices.pricing_level_id = 'default' THEN 3
                                ELSE 4
                            END,
                            llm_token_prices.effective_date DESC
                    ) AS rank
                FROM llm_usage
                LEFT JOIN llm_token_prices ON (llm_usage.llm_id = llm_token_prices.llm_id))
            -- Final Select, which just selects the correct fields, and adds a costs object
            SELECT
                id,
                started_at,
                if( any(llm_id) = 'weave_dummy_llm_id',
                any(summary_dump),
                concat(
                    left(any(summary_dump), length(any(summary_dump)) - 1),
                    ',"weave":{',
                        '"costs":',
                        concat(
                            '{',
                            arrayStringConcat(
                                groupUniqArray(
                                    concat(
                                        '"', toString(llm_id), '":{',
                                        '"prompt_tokens":', toString(prompt_tokens), ',',
                                        '"completion_tokens":', toString(completion_tokens), ',',
                                        '"requests":', toString(requests), ',',
                                        '"total_tokens":', toString(total_tokens), ',',
                                        '"prompt_token_cost":', toString(prompt_token_cost), ',',
                                        '"completion_token_cost":', toString(completion_token_cost), ',',
                                        '"prompt_tokens_total_cost":', toString(prompt_tokens * prompt_token_cost), ',',
                                        '"completion_tokens_total_cost":', toString(completion_tokens * completion_token_cost), ',',
                                        '"prompt_token_cost_unit":"', toString(prompt_token_cost_unit),  '",',
                                        '"completion_token_cost_unit":"', toString(completion_token_cost_unit),  '",',
                                        '"effective_date":"', toString(effective_date),  '",',
                                        '"provider_id":"', toString(provider_id),  '",',
                                        '"pricing_level":"', toString(pricing_level),  '",',
                                        '"pricing_level_id":"', toString(pricing_level_id),  '",',
                                        '"created_by":"', toString(created_by),  '",',
                                        '"created_at":"', toString(created_at),
                                    '"}'
                                    )
                                ), ','
                            ),
                            '} }'
                        ),
                    '}' )
                ) AS summary_dump
            FROM ranked_prices
            WHERE (rank = {pb_3:UInt64})
            GROUP BY id, started_at
        """,
        {
            "pb_0": ["a", "b"],
            "pb_1": "UHJvamVjdEludGVybmFsSWQ6Mzk1NDg2Mjc=",
            "pb_2": "UHJvamVjdEludGVybmFsSWQ6Mzk1NDg2Mjc=",
            "pb_3": 1,
        },
    )



import pytest

import weave
from weave.legacy.weave import box, weave_internal
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.artifact_fs import FilesystemArtifactFileType
from weave.legacy.weave.language_features.tagging import (
    make_tag_getter_op,
    tag_store,
    tagged_value_type,
    tagged_value_type_helpers,
)
from weave.legacy.weave.ops_domain.run_ops import run_tag_getter_op
from weave.legacy.weave.ops_domain.wb_domain_types import ProjectType, Run, RunType
from weave.legacy.weave.ops_primitives import dict as dict_ops
from weave.legacy.weave.ops_primitives import list_ as list_ops


def test_tagged_value():
    tv = tagged_value_type.TaggedValueType(
        types.TypedDict({"a": types.Number()}), types.String()
    )
    assert tv.value == types.String()
    assert tv.tag == types.TypedDict({"a": types.Number()})


def test_tagged_types():
    @weave.type()
    class _TestNumber:
        inner: int

    from weave.legacy.weave import context_state

    _loading_builtins_token = context_state.set_loading_built_ins()

    @weave.op()
    def add_tester(a: _TestNumber, b: _TestNumber) -> _TestNumber:
        return _TestNumber(a.inner + b.inner)

    @weave.op()
    def add_tester_2(d: _TestNumber, e: _TestNumber) -> _TestNumber:
        return _TestNumber(d.inner + e.inner)

    get_a_tag = make_tag_getter_op.make_tag_getter_op("a", _TestNumber.WeaveType())
    get_d_tag = make_tag_getter_op.make_tag_getter_op("d", _TestNumber.WeaveType())

    context_state.clear_loading_built_ins(_loading_builtins_token)

    # 1: Assert that that the tester works
    three = add_tester(_TestNumber(1), _TestNumber(2))
    assert weave.use(three).inner == 3

    # 2: Assert that we can get a tag
    assert weave.use(get_a_tag(three)).inner == 1

    # 3: Assert that we can use tagged values instead of raw values
    seven = add_tester(_TestNumber(3), _TestNumber(4))
    ten = add_tester_2(three, seven)
    assert tt(
        {"a": _TestNumber.WeaveType(), "d": _TestNumber.WeaveType()},
        _TestNumber.WeaveType(),
    ).assign_type(ten.type)
    assert (
        isinstance(ten.type, tagged_value_type.TaggedValueType)
        and isinstance(
            ten.type.tag.property_types["d"], tagged_value_type.TaggedValueType
        )
        and isinstance(ten.type.tag.property_types["d"].value, _TestNumber.WeaveType)
    )
    assert weave.use(ten).inner == 10

    # 4: Show that tags flow through
    assert weave.use(get_a_tag(ten)).inner == 1
    assert weave.use(get_d_tag(ten)).inner == 3

    # 5: Show that saving works:
    ten = weave.save(ten)
    assert weave.use(ten).inner == 10
    assert weave.use(get_a_tag(ten)).inner == 1


def test_index_checkpoint():
    arr = weave.save([1, 2, 3, 4])
    assert weave.use(arr) == [1, 2, 3, 4]
    tagged_arr = arr.createIndexCheckpointTag()
    assert weave.use(tagged_arr) == [1, 2, 3, 4]
    first_item = tagged_arr[0]
    assert weave.use(first_item) == 1
    first_item_index = first_item.indexCheckpoint()
    assert weave.use(first_item_index) == 0
    assert weave.use(tagged_arr[3].indexCheckpoint()) == 3


def tt(tag_dict, value_type):
    return tagged_value_type.TaggedValueType(types.TypedDict(tag_dict), value_type)


### Test Suite v2
@pytest.mark.parametrize(
    "target_type, next_type, is_assignable",
    [
        # empty tags and no tags
        (tt({}, types.Any()), types.Number(), False),
        (tt({}, types.Number()), types.Number(), False),
        (tt({}, types.String()), types.Number(), False),
        # no tags and empty tags
        (types.Any(), tt({}, types.Number()), True),
        (types.Number(), tt({}, types.Number()), True),
        (types.String(), tt({}, types.Number()), False),
        # no tags and some tags
        (types.Any(), tt({"a": types.Number()}, types.Number()), True),
        (types.Number(), tt({"a": types.Number()}, types.Number()), True),
        (types.String(), tt({"a": types.Number()}, types.Number()), False),
        # empty tags and empty tags
        (tt({}, types.Any()), tt({}, types.Number()), True),
        (tt({}, types.Number()), tt({}, types.Number()), True),
        (tt({}, types.String()), tt({}, types.Number()), False),
        # missing Tags
        (tt({"a": types.Number()}, types.Any()), tt({}, types.Number()), False),
        (tt({"a": types.Number()}, types.Number()), tt({}, types.Number()), False),
        (tt({"a": types.Number()}, types.String()), tt({}, types.Number()), False),
        # tag match
        (
            tt({"a": types.Number()}, types.Any()),
            tt({"a": types.Number()}, types.Number()),
            True,
        ),
        (
            tt({"a": types.Number()}, types.Number()),
            tt({"a": types.Number()}, types.Number()),
            True,
        ),
        (
            tt({"a": types.Number()}, types.String()),
            tt({"a": types.Number()}, types.Number()),
            False,
        ),
        # tag mismatch
        (
            tt({"a": types.Number()}, types.Any()),
            tt({"b": types.Number()}, types.Number()),
            False,
        ),
        (
            tt({"a": types.Number()}, types.Number()),
            tt({"b": types.Number()}, types.Number()),
            False,
        ),
        (
            tt({"a": types.Number()}, types.String()),
            tt({"b": types.Number()}, types.Number()),
            False,
        ),
        # tag superset
        (
            tt({"a": types.Number()}, types.Any()),
            tt({"a": types.Number(), "b": types.String()}, types.Number()),
            True,
        ),
        (
            tt({"a": types.Number()}, types.Number()),
            tt({"a": types.Number(), "b": types.String()}, types.Number()),
            True,
        ),
        (
            tt({"a": types.Number()}, types.String()),
            tt({"a": types.Number(), "b": types.String()}, types.Number()),
            False,
        ),
    ],
)
def test_tagged_value_assignment(target_type, next_type, is_assignable):
    assert target_type.assign_type(next_type) == is_assignable


def test_tag_lookups():
    @weave.type()
    class TestObj:
        inner: int

    # Test Simple Lookup
    obj_1 = TestObj(1)
    tag_store.add_tags(obj_1, {"a": 1})
    assert tag_store.find_tag(obj_1, "a") == 1

    obj_2 = TestObj(1)
    tag_store.add_tags(obj_2, {"a": 2, "b": 2})
    assert tag_store.find_tag(obj_2, "a") == 2
    assert tag_store.find_tag(obj_2, "b") == 2

    # Test Ancestor Lookup with name overlap
    obj_3 = TestObj(1)
    tag_store.add_tags(obj_3, {"nest": obj_2, "b": 3})
    assert tag_store.find_tag(obj_3, "a") == 2
    assert tag_store.find_tag(tag_store.find_tag(obj_3, "nest"), "b") == 2
    assert tag_store.find_tag(obj_3, "b") == 3


def test_tag_scope_with_multiple_children():
    list_node = weave.save([1, 2, 3])
    indexed = list_node.createIndexCheckpointTag()
    count_node = indexed.count()
    first_node = indexed[0]
    second_node = indexed[1]
    first_tag = first_node.indexCheckpoint()
    second_tag = second_node.indexCheckpoint()
    assert weave.use(count_node) == 3
    assert weave.use([first_tag, second_tag]) == [0, 1]


def test_nested_deserialization_of_weave0_tags():
    weave_0_tag_dict = {
        "type": "tagged",
        "tag": {
            "type": "union",
            "members": [
                {
                    "type": "tagged",
                    "tag": {
                        "type": "typedDict",
                        "propertyTypes": {"a": "string"},
                    },
                    "value": {
                        "type": "typedDict",
                        "propertyTypes": {"b": "number"},
                    },
                },
                {
                    "type": "tagged",
                    "tag": {
                        "type": "typedDict",
                        "propertyTypes": {"c": "boolean"},
                    },
                    "value": {
                        "type": "typedDict",
                        "propertyTypes": {"d": "project"},
                    },
                },
            ],
        },
        "value": {
            "type": "typedDict",
            "propertyTypes": {"e": "run"},
        },
    }
    weave_1_type = tagged_value_type.TaggedValueType.from_dict(weave_0_tag_dict)
    assert weave_1_type == tagged_value_type.TaggedValueType(
        types.TypedDict(
            {
                "a": types.optional(types.String()),
                "b": types.optional(types.Number()),
                "c": types.optional(types.Boolean()),
                "d": types.optional(ProjectType),
            }
        ),
        types.TypedDict(
            {
                "e": RunType,
            }
        ),
    )


def test_keytypes_tagged():
    object = box.box({"b": 1, "c": "a"})
    tag_store.add_tags(object, {"run": Run()})
    object_node = weave.save(object)

    kt_node = dict_ops.object_keytypes(object_node)
    assert weave.use(kt_node) == [
        {"key": "b", "type": types.Int()},
        {"key": "c", "type": types.String()},
    ]
    assert kt_node.type == tagged_value_type.TaggedValueType(
        types.TypedDict({"run": RunType.with_keys({})}),
        types.List(
            object_type=types.TypedDict(
                property_types={
                    "key": types.String(),
                    "type": types.TypeType(attr_types={}),
                }
            )
        ),
    )


@pytest.mark.parametrize(
    "list_data",
    [
        lambda: box.box([1, 2, 3]),
        lambda: box.box(
            [
                tag_store.add_tags(box.box(elem), {"int": i})
                for i, elem in enumerate([1, 2, 3])
            ]
        ),
    ],
)
def test_list_tags_accessible_to_map_elements(list_data):
    run = Run()
    tag_type = types.TypedDict({"run": RunType.with_keys({})})
    tagged = tag_store.add_tags(list_data(), {"run": run})
    saved_node = weave.save(tagged)
    map_fn = lambda row: run_tag_getter_op(row)
    fn = weave_internal.define_fn(
        {"row": tagged_value_type.TaggedValueType(tag_type, types.Int())}, map_fn
    )
    mapped = list_ops.List.map(saved_node, fn)
    assert weave.use(mapped) == [run, run, run]


def test_unwrap_rewrap_tags():
    t1 = tagged_value_type.TaggedValueType(
        types.TypedDict({"a": types.Int()}), types.Int()
    )

    unwrapped, rewrap = tagged_value_type_helpers.unwrap_tags(t1)
    assert unwrapped == types.Int()
    assert rewrap(unwrapped) == t1

    t2 = tagged_value_type.TaggedValueType(
        types.TypedDict({"a": types.Int()}), types.TypedDict({"b": types.Int()})
    )

    unwrapped, rewrap = tagged_value_type_helpers.unwrap_tags(t2)
    assert unwrapped == types.TypedDict({"b": types.Int()})
    assert rewrap(unwrapped) == t2

    t3 = tagged_value_type.TaggedValueType(
        types.TypedDict({"a": types.Int()}),
        tagged_value_type.TaggedValueType(
            types.TypedDict({"c": types.TypeType()}),
            types.TypedDict({"b": types.Int()}),
        ),
    )

    unwrapped, rewrap = tagged_value_type_helpers.unwrap_tags(t3)
    assert unwrapped == types.TypedDict({"b": types.Int()})
    assert rewrap(unwrapped) == t3

    t4 = types.TypedDict({"d": types.Int()})
    assert rewrap(t4) == tagged_value_type.TaggedValueType(
        types.TypedDict({"a": types.Int()}),
        tagged_value_type.TaggedValueType(
            types.TypedDict({"c": types.TypeType()}),
            types.TypedDict({"d": types.Int()}),
        ),
    )

    t5 = tagged_value_type.TaggedValueType(
        types.TypedDict({"a": types.Int()}),
        tagged_value_type.TaggedValueType(
            types.TypedDict({"c": types.TypeType()}),
            types.TypedDict({"d": types.List(types.String())}),
        ),
    )

    unwrapped, rewrap = tagged_value_type_helpers.unwrap_tags(t5)
    assert unwrapped == types.TypedDict({"d": types.List(types.String())})
    assert rewrap(unwrapped) == t5

    t6 = types.Int()
    unwrapped, rewrap = tagged_value_type_helpers.unwrap_tags(t6)
    assert unwrapped == t6
    assert rewrap(unwrapped) == t6



import contextlib
import contextvars
import copy
import dataclasses
import inspect
import random
import string
import typing

from weave.legacy.weave import weave_internal
from weave.legacy.weave import decorator_type, graph, ops, panel, weave_types
from weave.legacy.weave.language_features.tagging import tagged_value_type


@decorator_type.type()
class PanelDef:
    panelId: str
    panelVars: dict[str, graph.Node] = dataclasses.field(default_factory=dict)
    panelConfig: typing.Any = dataclasses.field(default_factory=lambda: None)


class SortDef(typing.TypedDict):
    dir: str
    columnId: str


_use_consistent_col_ids: contextvars.ContextVar[
    typing.Optional[bool]
] = contextvars.ContextVar("use_consistent_col_ids", default=False)


@contextlib.contextmanager
def use_consistent_col_ids(val: bool = True):
    token = _use_consistent_col_ids.set(val)
    yield
    _use_consistent_col_ids.reset(token)


@decorator_type.type()
class TableState:
    input_node: graph.Node = dataclasses.field(default_factory=lambda: graph.VoidNode())
    autoColumns: bool = dataclasses.field(default=False)
    columns: dict[str, PanelDef] = dataclasses.field(default_factory=dict)
    preFilterFunction: graph.Node = dataclasses.field(
        default_factory=lambda: graph.VoidNode()
    )
    columnNames: dict[str, str] = dataclasses.field(default_factory=dict)
    columnSelectFunctions: dict[str, graph.Node] = dataclasses.field(
        default_factory=dict
    )
    order: list[str] = dataclasses.field(default_factory=list)
    groupBy: list[str] = dataclasses.field(default_factory=list)
    sort: list[SortDef] = dataclasses.field(default_factory=list)
    pageSize: int = dataclasses.field(default=10)
    page: int = dataclasses.field(default=0)

    # def __init__(self, input_node):
    #     self._input_node = input_node
    #     self._auto_columns = False
    #     self._columns = {}
    #     self._pre_filter_function = graph.VoidNode()
    #     self._column_names = {}
    #     self._column_select_functions = {}
    #     self._order = []
    #     self._group_by = []
    #     self._sort = []
    #     self._page_size = 10
    #     self._page = 0

    def clone(self):
        return copy.deepcopy(self)

    def _new_col_id(self):
        if _use_consistent_col_ids.get():
            if hasattr(self, "_col_id_counter"):
                self._col_id_counter += 1
            else:
                self._col_id_counter = 0
            return f"col_{self._col_id_counter}"
        return "".join(
            random.choice(string.ascii_uppercase + string.digits) for _ in range(14)
        )

    def _expr_to_fn_node(self, fn, post_group_dimensionality=False):
        if isinstance(fn, graph.Node):
            return fn
        self_input_type = self.input_node.type
        if isinstance(self_input_type, weave_types.Function):
            self_input_type = self_input_type.output_type
        object_type = self_input_type.object_type

        # TODO: we are not deriving this type correctly
        if post_group_dimensionality:
            object_type = tagged_value_type.TaggedValueType(
                weave_types.TypedDict({}), weave_types.List(object_type)
            )

        sig = inspect.signature(fn)
        kwargs = {}
        if "domain" in sig.parameters:
            kwargs = {
                "domain": weave_internal.make_var_node(
                    weave_types.List(object_type), "domain"
                )
            }
        if "index" in sig.parameters:
            kwargs["index"] = weave_internal.make_var_node(weave_types.Int(), "index")
        return fn(weave_internal.make_var_node(object_type, "row"), **kwargs)

    def add_column(
        self,
        select_expr,
        name="",
        panel_def: typing.Optional[PanelDef] = None,
        groupby: bool = False,
        sort_dir: typing.Optional[str] = None,
    ):
        col_id = self._new_col_id()
        if panel_def is None:
            panel_def = PanelDef("", {}, None)
        self.columns[col_id] = panel_def
        self.columnNames[col_id] = name
        self.order.append(col_id)
        if groupby:
            self.enable_groupby(col_id)
        self.update_col(col_id, select_expr)
        if sort_dir:
            self.enable_sort(col_id, dir=sort_dir)
        return col_id

    def set_groupby(self, col_ids):
        self.group_by = col_ids

    def enable_groupby(self, col_id):
        if col_id not in self.groupBy:
            self.groupBy.append(col_id)

    def disable_groupby(self, col_id):
        if col_id in self.groupBy:
            self.groupBy.remove(col_id)

    def set_filter_fn(self, filter_exp):
        self.preFilterFunction = self._expr_to_fn_node(filter_exp)

    def enable_sort(self, col_id, dir="asc"):
        for sort_def in self.sort:
            if sort_def["columnId"] == col_id:
                sort_def["dir"] = dir
                return
        self.sort.append(SortDef(dir=dir, columnId=col_id))

    def update_col(self, col_id, select_expr):
        selected = self._expr_to_fn_node(
            select_expr, self.groupBy and col_id not in self.groupBy
        )

        if isinstance(selected, panel.Panel):
            self.columnSelectFunctions[col_id] = selected.input_node
            self.columns[col_id] = PanelDef(
                panelVars=selected.vars,
                panelId=selected.id,
                panelConfig=selected.config,
            )
        else:
            # TOTAL HACK HERE THIS IS NOT GENERAL!
            # TODO: abstract this behavior. We need to do this wherever we encounter a list
            # of Nodes. I think probably in lazy_call().
            if isinstance(selected, list):
                make_list_args = {}
                for i, item in enumerate(selected):
                    make_list_args["i%s" % i] = item
                selected = ops.make_list(**make_list_args)
            self.columnSelectFunctions[col_id] = selected

    def __eq__(self, other):
        return self.to_json() == other.to_json()

    def to_json(self):
        # TODO: its annoying that we have to manually rename everything to fix js
        # v python conventions. Automate or settle on one.
        columns = {
            id: {"panelId": v.panelId, "panelConfig": v.panelConfig}
            for (id, v) in self.columns.items()
        }
        pre_filter_function = self.preFilterFunction.to_json()
        column_select_functions = {
            id: v.to_json() for (id, v) in self.columnSelectFunctions.items()
        }
        return {
            "autoColumns": self.autoColumns,
            "columns": columns,
            "preFilterFunction": pre_filter_function,
            "columnNames": self.columnNames,
            "columnSelectFunctions": column_select_functions,
            "order": self.order,
            "groupBy": self.groupBy,
            "sort": self.sort,
            "pageSize": self.pageSize,
            "page": self.page,
        }



import dataclasses
import typing

import weave
from weave.legacy.weave import weave_internal
from weave.legacy.weave import graph, panel

ChoiceType = typing.TypeVar("ChoiceType")


@weave.type()
class ObjectPickerConfig(typing.Generic[ChoiceType]):
    label: str = dataclasses.field(default_factory=lambda: "")
    choice: graph.Node[ChoiceType] = dataclasses.field(default_factory=graph.VoidNode)


@weave.type()
class ObjectPicker(panel.Panel):
    id = "ObjectPicker"
    config: ObjectPickerConfig = dataclasses.field(default_factory=ObjectPickerConfig)

    def __init__(self, input_node=graph.VoidNode(), vars=None, config=None, **options):
        super().__init__(input_node=input_node, vars=vars)
        self.config = config
        if self.config is None:
            self.config = ObjectPickerConfig()
        if "label" in options:
            self.config.label = options["label"]

        # I originally tried to use a VarNode here. With the following comment:
        #   Set choice to a VarNode of the correct type (an item in our input
        #   list). This ensures that ObjectPickers stored in variables have a
        #   choice attribute that can be correctly dot-chained. Ie we want
        #   to maintain that the type is always correct.
        # self.config.choice = weave_internal.make_var_node(
        #     self.input_node.type.object_type, "choice"
        # )
        # That works on the Python side, but then we send the VarNode to the frontend
        # as the default value, and it sends it back down in the first execute request
        # for a panel that makes use of this panel's choice attribute.
        #
        # TODO: we probably need a way to reprent a void value that exists where a given
        # type is expected ("a hole"). A variable seems like a good way to do this.
        #
        # But a variable would mean we need to inspect objects' contents to determine
        # if something is executable. Because here we have Const nodes that contain
        # VarNodes in expressions in their values.
        if isinstance(self.config.choice, graph.VoidNode):
            self.config.choice = self.input_node[0]



# Two different implementations of an async map function that ensures
# no more than n tasks run in parallel. I did it two different ways to
# understand performance. Inconclusive. These are not used in the Weave
# code base right now.
# Exercise for read: Make these into async generators.

import asyncio
import typing


TaskInputType = typing.TypeVar("TaskInputType")
TaskOutputType = typing.TypeVar("TaskOutputType")


async def map_parallel_worker(
    task: typing.Callable[[TaskInputType], typing.Awaitable[TaskOutputType]],
    req_queue: asyncio.Queue[typing.Tuple[int, TaskInputType]],
    resp_queue: asyncio.Queue[typing.Tuple[int, TaskOutputType]],
) -> None:
    while True:
        item_index, item = await req_queue.get()
        resp = await task(item)
        await resp_queue.put((item_index, resp))


async def map_with_parallel_workers(
    items: typing.List[TaskInputType],
    task: typing.Callable[[TaskInputType], typing.Awaitable[TaskOutputType]],
    max_parallel: int = 100,
) -> typing.List[TaskOutputType]:
    req_queue: asyncio.Queue[typing.Tuple[int, TaskInputType]] = asyncio.Queue()
    resp_queue: asyncio.Queue[typing.Tuple[int, TaskOutputType]] = asyncio.Queue()

    resps: list[TaskOutputType] = [None] * len(items)  # type: ignore
    workers = []
    for i in range(max_parallel):
        workers.append(
            asyncio.create_task(map_parallel_worker(task, req_queue, resp_queue))
        )
    for i, item in enumerate(items):
        await req_queue.put((i, item))

    for i in range(len(items)):
        item_index, resp = await resp_queue.get()
        resps[item_index] = resp

    for worker in workers:
        worker.cancel()

    return resps


async def map_n_live_task_worker(
    task: typing.Callable[[TaskInputType], typing.Awaitable[TaskOutputType]],
    item_index: int,
    item: TaskInputType,
    resp_queue: asyncio.Queue[typing.Tuple[int, TaskOutputType]],
) -> None:
    resp = await task(item)
    await resp_queue.put((item_index, resp))


async def map_with_n_live_tasks(
    items: typing.List[TaskInputType],
    task: typing.Callable[[TaskInputType], typing.Awaitable[TaskOutputType]],
    max_parallel: int = 100,
) -> typing.List[TaskOutputType]:
    resp_queue: asyncio.Queue[typing.Tuple[int, TaskOutputType]] = asyncio.Queue()

    running = 0
    next_item_index = 0
    resps: list[TaskOutputType] = [None] * len(items)  # type: ignore
    while next_item_index < len(items) or running > 0:
        if running < max_parallel and next_item_index < len(items):
            item = items[next_item_index]
            asyncio.create_task(
                map_n_live_task_worker(task, next_item_index, item, resp_queue)
            )
            next_item_index += 1
            if next_item_index % 100 == 0:
                print("next_item_index", next_item_index, "running", running)
            running += 1
        else:
            idx, resp = await resp_queue.get()
            resps[idx] = resp
            running -= 1

    return resps



"""Utilities for working with IPython/Jupyter notebooks."""

import ast
from typing import Callable


class NotInteractiveEnvironmentError(Exception): ...


class ClassNotFoundError(ValueError): ...


def is_running_interactively() -> bool:
    """Check if the code is running in an interactive environment."""
    try:
        from IPython import get_ipython

        return get_ipython() is not None
    except ModuleNotFoundError:
        return False


def get_notebook_source() -> str:
    """Get the source code of the running notebook."""
    from IPython import get_ipython

    shell = get_ipython()
    if shell is None:
        raise NotInteractiveEnvironmentError

    if not hasattr(shell, "user_ns"):
        raise AttributeError("Cannot access user namespace")

    # This is the list of input cells in the notebook
    in_list = shell.user_ns["In"]

    # Stitch them back into a single "file"
    full_source = "\n\n".join(cell for cell in in_list[1:] if cell)

    return full_source


def get_class_source(cls: Callable) -> str:
    """Get the latest source definition of a class in the notebook."""
    notebook_source = get_notebook_source()
    tree = ast.parse(notebook_source)
    class_name = cls.__name__

    # We need to walk the entire tree and get the last one since that's the most version of the cls
    segment = None
    for node in ast.walk(tree):
        if isinstance(node, ast.ClassDef) and node.name == class_name:
            segment = ast.get_source_segment(notebook_source, node)

    if segment is not None:
        return segment

    raise ClassNotFoundError(f"Class '{class_name}' not found in the notebook")



import os

import weave
from weave.legacy.weave import weave_internal
from weave.legacy.weave import weave_internal as internal
from weave.legacy.weave import weave_types as types
from weave.legacy.weave import util, dispatch, graph
from weave.legacy.weave.panels import panel_board, panel_group, panel_trace
from weave.legacy.weave.panels.panel_trace_span import TraceSpanModelPanel, TraceSpanPanel
from weave.legacy.weave.panels_py import panel_autoboard
from weave.legacy.weave.panels_py.generator_templates import template_registry

panels = weave.legacy.weave.panels
ops = weave.legacy.weave.ops


# BOARD_ID must be unique across all ops. It must only contain letters and underscores.
BOARD_ID = "trace_monitor"

# BOARD_DISPLAY_NAME is the name that will be displayed in the UI
BOARD_DISPLAY_NAME = "Trace an LLM pipeline"

# BOARD_DESCRIPTION is the description that will be displayed in the UI
BOARD_DESCRIPTION = "Analyze traces from LLM chains, applications or any ML pipeline. Understand LLM agent behavior, monitor systems and debug ML pipelines."

# BOARD_INPUT_WEAVE_TYPE is the weave type of the input node.


BOARD_INPUT_WEAVE_TYPE = types.List(panel_trace.span_typed_dict_type)


board_name = "py_board-" + BOARD_ID


def make_span_table(input_node: graph.Node) -> panels.Table:
    traces_table = panels.Table(input_node)  # type: ignore
    traces_table.add_column(lambda row: row["status_code"].equal("SUCCESS"), "Success")
    traces_table.add_column(lambda row: row["name"], "Span Name")
    traces_table.add_column(lambda row: row["trace_id"], "Trace ID")
    traces_table.add_column(
        lambda row: row["end_time_s"] - row["start_time_s"], "Latency"
    )
    traces_table.add_column(lambda row: row["inputs"], "Inputs", panel_def="object")
    traces_table.add_column(lambda row: row["output"], "Output", panel_def="object")
    # traces_table.add_column(
    #     lambda row: row["attributes.model.id"], "Model ID", panel_def="object"
    # )
    traces_table.add_column(lambda row: row["timestamp"], "Timestamp", sort_dir="desc")

    return traces_table


@weave.op(  # type: ignore
    name=board_name,
    hidden=True,
    input_type={
        "input_node": types.Function(
            {},
            BOARD_INPUT_WEAVE_TYPE,
        )
    },
)
def board(
    input_node,  # type: ignore
) -> panels.Group:
    timestamp_col_name = "timestamp"

    ### Varbar

    # Add the input node as raw data
    varbar = panel_board.varbar(editable=False)

    all_spans = varbar.add("all_spans", input_node)

    trace_roots = all_spans.filter(lambda row: row["parent_id"].isNone())

    trace_roots_var = varbar.add("trace_roots", trace_roots, True)

    filter_fn = varbar.add(
        "filter_fn",
        weave_internal.define_fn(
            {"row": trace_roots.type.object_type},
            lambda row: weave_internal.const(True),
        ),
        hidden=True,
    )

    grouping_fn = varbar.add(
        "grouping_fn",
        weave_internal.define_fn(
            {"row": input_node.type.object_type}, lambda row: row["name"]
        ),
        hidden=True,
    )

    filtered_data = varbar.add(
        "filtered_data", trace_roots.filter(filter_fn), hidden=True
    )

    # Setup date range variables:
    ## 1. raw_data_range is derived from raw_data
    filtered_range = varbar.add(
        "filtered_range",
        weave.legacy.weave.ops.make_list(
            a=filtered_data[timestamp_col_name].min(),
            b=filtered_data[timestamp_col_name].max(),
        ),
        hidden=True,
    )

    ## 2. user_zoom_range is used to store the user's zoom range
    user_zoom_range = varbar.add("user_zoom_range", None, hidden=True)

    ## 2.b: Setup a date picker to set the user_zoom_range
    varbar.add(
        "time_range",
        weave.legacy.weave.panels.DateRange(
            user_zoom_range, domain=trace_roots[timestamp_col_name]
        ),
    )

    ## 3. bin_range is derived from user_zoom_range and raw_data_range. This is
    ##    the range of data that will be displayed in the charts.
    bin_range = varbar.add(
        "bin_range", user_zoom_range.coalesce(filtered_range), hidden=True
    )
    # Derive the windowed data to use in the plots as a function of bin_range

    window_data = varbar.add(
        "window_data",
        trace_roots.filter(
            lambda row: weave.legacy.weave.ops.Boolean.bool_and(
                row[timestamp_col_name] >= bin_range[0],
                row[timestamp_col_name] <= bin_range[1],
            )
        ),
        hidden=True,
    )

    filters = varbar.add(
        "filters",
        weave.legacy.weave.panels.FilterEditor(filter_fn, node=window_data),
    )

    filtered_window_data = varbar.add(
        "filtered_window_data", window_data.filter(filter_fn), hidden=True
    )

    ### Overview tab

    overview_tab = weave.legacy.weave.panels.Group(
        layoutMode="grid",
        showExpressions=True,
        enableAddPanel=True,
    )

    # traces_table =   # type: ignore

    overview_tab.add(
        "latency_over_time",
        panel_autoboard.timeseries(
            filtered_data,
            bin_domain_node=bin_range,
            x_axis_key="timestamp",
            y_expr=lambda row: row.map(
                lambda ir: ir["end_time_s"] - ir["start_time_s"]
            ).avg(),
            y_title="avg latency (s)",
            color_expr=lambda row: grouping_fn(row),
            color_title="Root Span Name",
            x_domain=user_zoom_range,
            n_bins=50,
        ),
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=0, y=0, w=6, h=6),
    )

    overview_tab.add(
        "latency_distribution",
        filtered_window_data.map(lambda row: row["end_time_s"] - row["start_time_s"]),
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=6, y=0, w=6, h=6),
    )

    overview_tab.add(
        "success_over_time",
        panel_autoboard.timeseries(
            filtered_data,
            bin_domain_node=bin_range,
            x_axis_key="timestamp",
            y_expr=lambda row: (
                row.filter(
                    lambda inner_row: inner_row["status_code"] == "SUCCESS"
                ).count()
                / row.count()
            ),
            y_title="success rate",
            color_expr=lambda row: grouping_fn(row),
            color_title="Root Span Name",
            x_domain=user_zoom_range,
            n_bins=50,
        ),
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=12, y=0, w=6, h=6),
    )

    overview_tab.add(
        "success_distribution",
        weave.legacy.weave.ops.dict_(
            **{
                "success": filtered_window_data.filter(
                    lambda row: row["status_code"] == "SUCCESS"
                ).count(),
                "error": filtered_window_data.filter(
                    lambda row: row["status_code"] != "SUCCESS"
                ).count(),
            }
        ),
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=18, y=0, w=6, h=6),
    )

    traces_table_var = overview_tab.add(
        "traces_table",
        make_span_table(filtered_window_data),
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=0, y=6, w=24, h=6),
    )

    trace_spans = all_spans.filter(
        lambda row: row["trace_id"] == traces_table_var.active_data()["trace_id"]
    )

    trace_viewer = panels.Trace(trace_spans)  # type: ignore

    trace_viewer_var = overview_tab.add(
        "trace_viewer",
        trace_viewer,
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=0, y=12, w=16, h=6),
    )

    selected_trace_model = overview_tab.add(
        "selected_trace_model",
        TraceSpanModelPanel(traces_table_var.active_data()),
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=0, y=18, w=16, h=6),
    )

    active_span = trace_viewer_var.active_span()

    selected_span_details = overview_tab.add(
        "selected_span_details",
        TraceSpanPanel(active_span),
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=16, y=12, w=8, h=12),
    )

    similar_spans = all_spans.filter(lambda row: row["name"] == active_span["name"])

    similar_spans_table = make_span_table(similar_spans)
    similar_spans_table_var = overview_tab.add(
        "similar_spans_table",
        similar_spans_table,
        layout=weave.legacy.weave.panels.GroupPanelLayout(x=0, y=22, w=24, h=6),
    )

    return panels.Board(vars=varbar, panels=overview_tab)


instructions_md = util.read_or_default(
    os.path.join(os.path.dirname(__file__), "instructions", "panel_trace_monitor.md")
)

template_registry.register(
    board_name,
    BOARD_DISPLAY_NAME,
    BOARD_DESCRIPTION,
    is_featured=True,
    instructions_md=instructions_md,
    thumbnail_url="https://raw.githubusercontent.com/wandb/weave/master/docs/assets/trace-an-llm-pipeline.png",
)



import typing

from weave.legacy.weave import weave_types as types
from weave.legacy.weave.language_features.tagging import tagged_value_type

if typing.TYPE_CHECKING:
    from weave.legacy.weave import op_def as OpDef


# Determines if the op_def should ftag the output with the input. Currently, this is true
# as long as the op does not consume tags and the input of the op is an ObjectType. This
# may evolve over time.
def should_tag_op_def_outputs(op_def: "OpDef.OpDef") -> bool:
    named_args = op_def.input_type.named_args()
    return (
        should_flow_tags(op_def)
        # NOTICE: For now, we only autotag with sublcasses of ObjectType that
        # are named Test*, plus ProjectType and RunType.
        # Uncomment this during advanced testing to ensure tags work in the general cases
        and (
            (
                isinstance(named_args[0].type, types.ObjectType)
                and named_args[0].type.name.startswith("_Test")
            )
            or named_args[0].type.__class__.__name__ in ["projectType", "runType"]
        )
    )


# Determines if the op_def should flow tags from the input to the output. Currently,
# we assume yes, unless the op_def consumes a tag. This may evolve over time
def should_flow_tags(op_def: "OpDef.OpDef") -> bool:
    named_args = op_def.input_type.named_args()
    return (
        not op_def_consumes_tags(op_def)
        and len(named_args) > 0
        and not isinstance(named_args[0].type, types.Function)
        and not op_def.name.endswith("dropTags")
    )


# Helper function to determine if the op_def consumes a tag
def op_def_consumes_tags(op_def: "OpDef.OpDef") -> bool:
    named_args = op_def.input_type.named_args()
    return len(named_args) > 0 and tagged_value_type.TaggedValueType(
        types.TypedDict({}), types.Any()
    ).assign_type(named_args[0].type)


# Helper function to get the first argument of the op_def
def get_first_arg(
    op_def: "OpDef.OpDef", args: list[typing.Any], kwargs: dict[str, typing.Any]
) -> typing.Tuple[str, typing.Any]:
    params = op_def.input_type.create_param_dict(args, kwargs)
    key = list(params.keys())[0]
    return (key, params[key])



import contextlib
import logging
import os
import typing
from urllib.parse import urlparse

from weave.legacy.weave import util
from weave.legacy.weave import urls, client, context_state
from weave.legacy.weave.client_interface import ClientInterface


@contextlib.contextmanager
def execution_client():
    """Returns a client for use by the execution engine and op resolvers."""
    from weave.legacy.weave import server

    # Force in process execution
    with context_state.client(client.NonCachingClient(server.InProcessServer())):
        with context_state.analytics_disabled():
            yield


@contextlib.contextmanager
def local_http_client():
    from weave.legacy.weave import server

    s = server.HttpServer()
    s.start()
    with context_state.server(s):
        with context_state.client(server.HttpServerClient(s.url)):
            yield
    s.shutdown()


@contextlib.contextmanager
def weavejs_client():
    from weave.legacy.weave import server

    s = server.HttpServer()
    s.start()
    with context_state.server(s):
        with context_state.client(server.HttpServerClient(s.url, emulate_weavejs=True)):
            yield


def use_fixed_server_port():
    """Force Weave server to port 9994 so wandb frontend can talk to it."""
    # s = server.HttpServer(port=9994)
    # s.start()
    # _weave_client.set(server.HttpServerClient(s.url))
    from weave.legacy.weave import server

    context_state.set_client(server.HttpServerClient("http://localhost:9994"))


def use_frontend_devmode():
    """Talk to external server running on 9994"""
    use_fixed_server_port()
    urls.use_local_urls()

    # point frontend to vite server
    context_state.set_frontend_url("http://localhost:3000")


def use_lazy_execution():
    context_state._eager_mode.set(False)


lazy_execution = context_state.lazy_execution


def _make_default_client():
    from weave.legacy.weave import server

    if util.is_notebook():
        serv = context_state.get_server()
        if serv is None:
            frontend_url = os.environ.get("WEAVE_FRONTEND_URL")
            if frontend_url is not None:
                parsed_url = urlparse(frontend_url)
                server_args = {}
                if parsed_url.hostname is not None:
                    server_args["host"] = parsed_url.hostname
                if parsed_url.port is not None:
                    server_args["port"] = parsed_url.port
                serv = server.HttpServer(**server_args)
            else:
                serv = server.HttpServer()
            serv.start()
            context_state.set_server(serv)
        # Falling through here means the notebook kernel uses
        # InprocessServer, but the frontend uses HttpServer.
        # versions() doesn't work when we use the HttpServer currently.
        # return server.HttpServerClient(serv.url)

    return client.Client(server.InProcessServer())


def get_client() -> typing.Optional[ClientInterface]:
    c = context_state.get_client()
    if c is None:
        c = _make_default_client()
        context_state.set_client(c)
    return c


def get_frontend_url():
    from weave.legacy.weave import server

    url = os.environ.get("WEAVE_FRONTEND_URL", context_state.get_frontend_url())
    if url is None:
        client = get_client()
        if isinstance(client, server.HttpServerClient):
            url = client.url
        else:
            s = context_state.get_server()
            if s is not None:
                url = s.url
            else:
                raise RuntimeError("Frontend server is not running")
    url += "/__frontend/weave_jupyter"
    return url
